import datetime
import logging
import time
import warnings

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.mail import mail_admins
from django.db.models import F, Q
from django.utils import timezone

from .base.models import AbstractRecurringUserPlan
from .signals import account_automatic_renewal

User = get_user_model()
logger = logging.getLogger("plans.tasks")


def get_active_plans():
    return (
        User.objects.select_related("userplan")
        .filter(userplan__active=True)
        .exclude(userplan__expire=None)
    )


def autorenew_account(
    providers=None, throttle_seconds=0, catch_exceptions=False, dry_run=False
):
    logger.info("Started automatic account renewal")
    PLANS_AUTORENEW_SCHEDULE = getattr(settings, "PLANS_AUTORENEW_SCHEDULE", None)
    PLANS_AUTORENEW_BEFORE_DAYS = getattr(settings, "PLANS_AUTORENEW_BEFORE_DAYS", 0)
    PLANS_AUTORENEW_BEFORE_HOURS = getattr(settings, "PLANS_AUTORENEW_BEFORE_HOURS", 0)

    accounts_to_check = User.objects.select_related(
        "userplan", "userplan__recurring"
    ).filter(
        userplan__recurring__renewal_triggered_by=AbstractRecurringUserPlan.RENEWAL_TRIGGERED_BY.TASK,
        userplan__recurring__token_verified=True,
    )

    if PLANS_AUTORENEW_SCHEDULE is not None:
        if PLANS_AUTORENEW_BEFORE_DAYS or PLANS_AUTORENEW_BEFORE_HOURS:
            logger.warning(
                "PLANS_AUTORENEW_SCHEDULE is set, ignoring PLANS_AUTORENEW_BEFORE_DAYS and PLANS_AUTORENEW_BEFORE_HOURS"
            )
        now_dt = timezone.now()
        q = Q()
        max_renew_after = getattr(
            settings,
            "PLANS_AUTORENEW_MAX_DAYS_AFTER_EXPIRY",
            datetime.timedelta(days=30),
        )
        for schedule in PLANS_AUTORENEW_SCHEDULE:
            q |= Q(
                Q(userplan__recurring__last_renewal_attempt__isnull=True)
                | Q(
                    userplan__recurring__last_renewal_attempt__lt=F("userplan__expire")
                    - schedule
                ),
                userplan__expire__lte=now_dt + schedule,
                userplan__expire__gte=now_dt + schedule - max_renew_after,
            )
        accounts_for_renewal = accounts_to_check.filter(q).distinct()
    else:
        warnings.warn(
            "PLANS_AUTORENEW_BEFORE_DAYS and PLANS_AUTORENEW_BEFORE_HOURS are deprecated "
            "and will be removed in a future version. "
            "Please use PLANS_AUTORENEW_SCHEDULE instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        accounts_for_renewal = accounts_to_check.filter(
            userplan__expire__lt=timezone.now()
            + datetime.timedelta(
                days=PLANS_AUTORENEW_BEFORE_DAYS, hours=PLANS_AUTORENEW_BEFORE_HOURS
            ),
        )

    if providers:
        accounts_for_renewal = accounts_for_renewal.filter(
            userplan__recurring__payment_provider__in=providers
        )

    logger.info(f"{accounts_for_renewal.count()} accounts to be renewed.")

    accounts_for_renewal = accounts_for_renewal.all()

    if dry_run:
        logger.info("Dry run mode: No changes will be made.")
        for user in accounts_for_renewal:
            logger.info(f"DRY RUN: Would renew user {user.pk} ({user.email})")
            if hasattr(user, "userplan") and not user.userplan.is_active():
                logger.info(
                    f"DRY RUN: Would activate userplan for user {user.pk} ({user.email})"
                )
            logger.info(
                f"DRY RUN: Would send account_automatic_renewal signal for user {user.pk} ({user.email})"
            )
        return accounts_for_renewal

    renewed_accounts = []
    for user in accounts_for_renewal:
        if hasattr(user, "userplan") and hasattr(user.userplan, "recurring"):
            user.userplan.recurring.last_renewal_attempt = timezone.now()
            user.userplan.recurring.save(update_fields=["last_renewal_attempt"])
        if throttle_seconds:
            time.sleep(throttle_seconds)
        if catch_exceptions:
            try:
                account_automatic_renewal.send(sender=None, user=user)
            except Exception as e:
                logger.error(
                    f"Error renewing account for user {user.pk} ({user.email}): {e}",
                    exc_info=True,
                )
                subject = f"Failed to renew account for user {user.pk} ({user.email})"
                message = f"""
                An error occurred while trying to automatically renew the account for user:
                User ID: {user.pk}
                User email: {user.email}

                Error details:
                {e}
                """
                mail_admins(subject, message, fail_silently=True)
        else:
            account_automatic_renewal.send(sender=None, user=user)
        renewed_accounts.append(user)
    return renewed_accounts


def expire_account():
    logger.info("Started account expiration")

    expired_accounts = get_active_plans().filter(
        userplan__expire__lt=datetime.date.today()
    )

    for user in expired_accounts.all():
        user.userplan.expire_account()

    notifications_days_before = getattr(settings, "PLANS_EXPIRATION_REMIND", [])

    if notifications_days_before:
        days = map(
            lambda x: datetime.date.today() + datetime.timedelta(days=x),
            notifications_days_before,
        )
        for user in User.objects.select_related("userplan").filter(
            userplan__active=True, userplan__expire__in=days
        ):
            user.userplan.remind_expire_soon()



from __future__ import unicode_literals

import logging
import re
import warnings
from datetime import date, timedelta
from decimal import Decimal

import stdnum.eu.vat
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models, transaction

from plans import utils

try:
    from django.contrib.sites.models import Site
except RuntimeError:
    Site = None
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.template import Context
from django.template.base import Template
from django.urls import reverse
from django.utils import translation
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from django_countries.fields import CountryField
from ordered_model.models import OrderedModel
from sequences import get_next_value
from swapper import load_model

from plans.contrib import get_user_language, send_template_email
from plans.enumeration import Enumeration
from plans.signals import (
    account_activated,
    account_change_plan,
    account_deactivated,
    account_expired,
    order_completed,
)
from plans.taxation.eu import EUTaxationPolicy
from plans.utils import country_code_transform, get_country_code, get_currency
from plans.validators import plan_validation

accounts_logger = logging.getLogger("accounts")


class BaseMixin(models.Model):
    created = models.DateTimeField(
        _("created"), db_index=True, auto_now_add=True, null=True
    )
    updated_at = models.DateTimeField(auto_now=True, null=True)

    class Meta:
        abstract = True

    @classmethod
    def get_concrete_model(cls):
        return load_model("plans", cls.__name__.replace("Abstract", ""))


class AbstractPlan(BaseMixin, OrderedModel):
    """
    Single plan defined in the system. A plan can customized (referred to user) which means
    that only this user can purchase this plan and have it selected.

    Plan also can be visible and available. Plan is displayed on the list of currently available plans
    for user if it is visible. User cannot change plan to a plan that is not visible. Available means
    that user can buy a plan. If plan is not visible but still available it means that user which
    is using this plan already will be able to extend this plan again. If plan is not visible and not
    available, he will be forced then to change plan next time he extends an account.
    """

    name = models.CharField(_("name"), max_length=100)
    slug = models.SlugField(_("slug"), max_length=100, unique=True)
    description = models.TextField(_("description"), blank=True)
    default = models.BooleanField(
        help_text=_('Both "Unknown" and "No" means that the plan is not default'),
        default=None,
        db_index=True,
        unique=True,
        null=True,
    )
    available = models.BooleanField(
        _("available"),
        default=False,
        db_index=True,
        help_text=_("Is still available for purchase"),
    )
    visible = models.BooleanField(
        _("visible"),
        default=True,
        db_index=True,
        help_text=_("Is visible in current offer"),
    )
    customized = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        verbose_name=_("customized"),
        on_delete=models.CASCADE,
    )
    quotas = models.ManyToManyField("Quota", through="PlanQuota")
    url = models.URLField(
        max_length=200,
        blank=True,
        help_text=_(
            "Optional link to page with more information (for clickable pricing table headers)"
        ),
    )

    class Meta:
        abstract = True
        ordering = ("order",)
        verbose_name = _("Plan")
        verbose_name_plural = _("Plans")

    @classmethod
    def get_default_plan(cls):
        try:
            return_value = cls.objects.get(default=True)
        except cls.DoesNotExist:
            return_value = None
        return return_value

    @classmethod
    def get_current_plan(cls, user):
        """Get current plan for user. If userplan is expired, get default plan"""
        if (
            not user
            or user.is_anonymous
            or not hasattr(user, "userplan")
            or user.userplan.is_expired()
        ):
            default_plan = cls.get_default_plan()
            if default_plan is None or not default_plan.is_free():
                raise ValidationError(_("User plan has expired"))
            return default_plan
        return user.userplan.plan

    def __str__(self):
        return self.name

    def get_quota_dict(self):
        return dict(self.planquota_set.values_list("quota__codename", "value"))

    def is_free(self):
        return self.planpricing_set.count() == 0

    is_free.boolean = True


class AbstractBillingInfo(BaseMixin, models.Model):
    """
    Stores customer billing data needed to issue an invoice
    """

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, verbose_name=_("user"), on_delete=models.CASCADE
    )
    tax_number = models.CharField(
        _("VAT ID"), max_length=200, blank=True, db_index=True
    )
    name = models.CharField(_("name"), max_length=200, db_index=True)
    street = models.CharField(_("street"), max_length=200)
    zipcode = models.CharField(_("zip code"), max_length=200)
    city = models.CharField(_("city"), max_length=200)
    country = CountryField(_("country"))
    shipping_name = models.CharField(
        _("name (shipping)"), max_length=200, blank=True, help_text=_("optional")
    )
    shipping_street = models.CharField(
        _("street (shipping)"), max_length=200, blank=True, help_text=_("optional")
    )
    shipping_zipcode = models.CharField(
        _("zip code (shipping)"), max_length=200, blank=True, help_text=_("optional")
    )
    shipping_city = models.CharField(
        _("city (shipping)"), max_length=200, blank=True, help_text=_("optional")
    )

    class Meta:
        abstract = True
        verbose_name = _("Billing info")
        verbose_name_plural = _("Billing infos")

    @staticmethod
    def get_full_tax_number(tax_number, country):
        number = tax_number
        if tax_number.startswith(country):
            number = tax_number[len(country) :]
        return country_code_transform(country) + number

    @staticmethod
    def clean_tax_number(tax_number, country):
        tax_number = re.sub(r"[^A-Z0-9]", "", tax_number.upper())

        country_str = tax_number[: len(country)]
        if country_str == country_code_transform(country):
            country = country_code_transform(country)

        if country_str.isalpha() and country_str != country:
            raise ValidationError(
                _("VAT ID country code doesn't corespond with country")
            )

        if tax_number and country:
            if country.lower() in stdnum.eu.vat.MEMBER_STATES:
                full_number = (
                    AbstractBillingInfo.get_concrete_model().get_full_tax_number(
                        tax_number, country
                    )
                )
                try:
                    return stdnum.eu.vat.validate(full_number)
                except stdnum.exceptions.ValidationError as e:
                    raise ValidationError(_(f"VAT ID is not correct: {e.message}"))

            return tax_number
        else:
            return ""


class AbstractUserPlan(BaseMixin, models.Model):
    """
    Currently selected plan for user account.
    """

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, verbose_name=_("user"), on_delete=models.CASCADE
    )
    plan = models.ForeignKey("Plan", verbose_name=_("plan"), on_delete=models.CASCADE)
    expire = models.DateField(
        _("expire"), default=None, blank=True, null=True, db_index=True
    )
    active = models.BooleanField(_("active"), default=True, db_index=True)

    class Meta:
        abstract = True
        verbose_name = _("User plan")
        verbose_name_plural = _("Users plans")

    def __str__(self):
        return "%s [%s]" % (self.user, self.plan)

    def is_active(self):
        return self.active

    def is_expired(self):
        if self.expire is None:
            return False
        else:
            return self.expire < date.today()

    def days_left(self):
        if self.expire is None:
            return None
        else:
            return (self.expire - date.today()).days

    def clean_activation(self):
        errors = plan_validation(self.user)
        if not errors["required_to_activate"]:
            plan_validation(self.user, on_activation=True)
            self.activate()
        else:
            self.deactivate()
        return errors

    def activate(self):
        if not self.active:
            self.active = True
            self.save()
            account_activated.send(sender=self, user=self.user)

    def deactivate(self):
        if self.active:
            self.active = False
            self.save()
            account_deactivated.send(sender=self, user=self.user)

    def initialize(self):
        """
        Set up user plan for first use
        """
        if not self.is_active():
            # Plans without pricings don't need to expire
            if self.expire is None and self.plan.planpricing_set.count():
                self.expire = now() + timedelta(
                    days=getattr(settings, "PLANS_DEFAULT_GRACE_PERIOD", 30)
                )
            self.activate()  # this will call self.save()

    def get_plan_extended_from(self, plan):
        if plan.is_free():
            return None
        if not self.is_expired() and self.expire is not None and self.plan == plan:
            return self.expire
        return date.today()

    def has_automatic_renewal(self):
        return (
            hasattr(self, "recurring")
            and self.recurring.renewal_triggered_by
            != self.recurring.RENEWAL_TRIGGERED_BY.USER
            and self.recurring.token_verified
        )

    def get_plan_extended_until(self, plan, pricing):
        if plan.is_free():
            return None
        if pricing is None:
            return self.expire
        return self.get_plan_extended_from(plan) + timedelta(days=pricing.period)

    def get_plan_reduced_until(self, pricing):
        if self.expire is None:
            return self.expire
        if pricing is None:
            return self.expire
        return self.expire - timedelta(days=pricing.period)

    def plan_autorenew_at(self):
        """
        Helper function which calculates when the plan autorenewal will occur
        """
        if self.expire:
            plans_autorenew_before_days = getattr(
                settings, "PLANS_AUTORENEW_BEFORE_DAYS", 0
            )
            plans_autorenew_before_hours = getattr(
                settings, "PLANS_AUTORENEW_BEFORE_HOURS", 0
            )
            return self.expire - timedelta(
                days=plans_autorenew_before_days, hours=plans_autorenew_before_hours
            )

    def set_plan_renewal(
        self,
        order,
        # TODO: has_automatic_renewal deprecated. Remove in the next major release.
        has_automatic_renewal=None,
        # TODO: renewal_triggered_by=None deprecated. Set to TASK in the next major release.
        renewal_triggered_by=None,
        **kwargs,
    ):
        """
        Creates or updates plan renewal information for this userplan with given order
        """
        if not hasattr(self, "recurring"):
            self.recurring = AbstractRecurringUserPlan.get_concrete_model()()

        if has_automatic_renewal is None and renewal_triggered_by is None:
            has_automatic_renewal = True
        if has_automatic_renewal is not None:
            warnings.warn(
                "has_automatic_renewal is deprecated. Use renewal_triggered_by instead.",
                DeprecationWarning,
            )
        if renewal_triggered_by is None:
            warnings.warn(
                "renewal_triggered_by=None is deprecated. "
                "Set an AbstractRecurringUserPlan.RENEWAL_TRIGGERED_BY instead.",
                DeprecationWarning,
            )
            renewal_triggered_by = (
                self.recurring.RENEWAL_TRIGGERED_BY.TASK
                if has_automatic_renewal
                else self.recurring.RENEWAL_TRIGGERED_BY.USER
            )

        # Erase values of all fields
        # We don't want to mix the old and new values
        self.recurring.set_all_fields_default()

        # Set new values
        self.recurring.user_plan = self
        self.recurring.pricing = order.pricing
        self.recurring.amount = order.amount
        self.recurring.tax = order.tax
        self.recurring.currency = order.currency
        self.recurring.renewal_triggered_by = renewal_triggered_by
        for k, v in kwargs.items():
            setattr(self.recurring, k, v)
        self.recurring.save()
        return self.recurring

    def extend_account(self, plan, pricing):
        """
        Manages extending account after plan or pricing order
        :param plan:
        :param pricing: if pricing is None then account will be only upgraded
        :return:
        """

        status = False  # flag; if extending account was successful?
        expire = self.get_plan_extended_until(plan, pricing)
        if pricing is None:
            # Process a plan change request (downgrade or upgrade)
            # No account activation or extending at this point
            self.plan = plan

            if self.expire is not None and not plan.planpricing_set.count():
                # Assume no expiry date for plans without pricing.
                self.expire = None

            self.save()
            account_change_plan.send(sender=self, user=self.user)
            if getattr(settings, "PLANS_SEND_EMAILS_PLAN_CHANGED", True):
                mail_context = {"user": self.user, "userplan": self, "plan": plan}
                send_template_email(
                    [self.user.email],
                    "mail/change_plan_title.txt",
                    "mail/change_plan_body.txt",
                    mail_context,
                    get_user_language(self.user),
                )
            accounts_logger.info(
                "Account '%s' [id=%d] plan changed to '%s' [id=%d]"
                % (self.user, self.user.pk, plan, plan.pk)
            )
            status = True
        else:
            # Processing standard account extending procedure
            if self.plan == plan:
                status = True
            else:
                # This should not ever happen (as this case should be managed by plan change request)
                # but just in case we consider a case when user has a different plan
                if not self.plan.is_free() and self.expire is None:
                    status = True
                elif not self.plan.is_free() and self.expire > date.today():
                    status = False
                    accounts_logger.warning(
                        "Account '%s' [id=%d] plan NOT changed to '%s' [id=%d]"
                        % (self.user, self.user.pk, plan, plan.pk)
                    )
                else:
                    status = True
                    account_change_plan.send(sender=self, user=self.user)
                    self.plan = plan

            if status:
                self.expire = expire
                self.save()
                accounts_logger.info(
                    "Account '%s' [id=%d] has been extended by %d days using plan '%s' [id=%d]"
                    % (self.user, self.user.pk, pricing.period, plan, plan.pk)
                )
                if getattr(settings, "PLANS_SEND_EMAILS_PLAN_EXTENDED", True):
                    mail_context = {
                        "user": self.user,
                        "userplan": self,
                        "plan": plan,
                        "pricing": pricing,
                    }
                    send_template_email(
                        [self.user.email],
                        "mail/extend_account_title.txt",
                        "mail/extend_account_body.txt",
                        mail_context,
                        get_user_language(self.user),
                    )

        if status:
            self.clean_activation()

        return status

    def reduce_account(self, pricing):
        """
        Manages reducing account after returning an order
        :param pricing: if pricing is None then nothing is changed
        :return:
        """
        if pricing is not None:
            self.expire = self.get_plan_reduced_until(pricing)
            self.save()

    def expire_account(self):
        """manages account expiration"""

        self.deactivate()

        accounts_logger.info(
            "Account '%s' [id=%d] has expired" % (self.user, self.user.pk)
        )

        mail_context = {"user": self.user, "userplan": self}
        send_template_email(
            [self.user.email],
            "mail/expired_account_title.txt",
            "mail/expired_account_body.txt",
            mail_context,
            get_user_language(self.user),
        )

        account_expired.send(sender=self, user=self.user)

    def remind_expire_soon(self):
        """reminds about soon account expiration"""

        mail_context = {"user": self.user, "userplan": self, "days": self.days_left()}
        send_template_email(
            [self.user.email],
            "mail/remind_expire_title.txt",
            "mail/remind_expire_body.txt",
            mail_context,
            get_user_language(self.user),
        )

    @classmethod
    def create_for_user(cls, user):
        default_plan = AbstractPlan.get_concrete_model().get_default_plan()
        if default_plan is not None:
            UserPlan = AbstractUserPlan.get_concrete_model()
            return UserPlan.objects.create(
                user=user,
                plan=default_plan,
                active=False,
                expire=None,
            )

    @classmethod
    def create_for_users_without_plan(cls):
        userplans = get_user_model().objects.filter(userplan=None)
        for user in userplans:
            AbstractUserPlan.get_concrete_model().create_for_user(user)
        return userplans

    def get_current_plan(self):
        """Tiny helper, very usefull in templates"""
        return AbstractPlan.get_concrete_model().get_current_plan(self.user)


class AbstractRecurringUserPlan(BaseMixin, models.Model):
    """
    OneToOne model associated with UserPlan that stores information about the plan recurrence.
    More about recurring payments in docs.
    """

    RENEWAL_TRIGGERED_BY = Enumeration(
        [
            (1, "OTHER", pgettext_lazy("Renewal triggered by", "other")),
            (2, "USER", pgettext_lazy("Renewal triggered by", "user")),
            (3, "TASK", pgettext_lazy("Renewal triggered by", "task")),
        ]
    )

    user_plan = models.OneToOneField(
        "UserPlan", on_delete=models.CASCADE, related_name="recurring"
    )
    token = models.CharField(
        _("recurring token"),
        help_text=_(
            "Token, that will be used for payment renewal. Depends on used payment provider"
        ),
        max_length=255,
        default=None,
        null=True,
        blank=True,
    )
    payment_provider = models.CharField(
        _("payment provider"),
        help_text=_("Provider, that will be used for payment renewal"),
        max_length=255,
        default=None,
        null=True,
        blank=True,
    )
    pricing = models.ForeignKey(
        "Pricing",
        help_text=_("Recurring pricing"),
        default=None,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    amount = models.DecimalField(
        _("amount"),
        max_digits=7,
        decimal_places=2,
        db_index=True,
        null=True,
        blank=True,
    )
    tax = models.DecimalField(
        _("tax"), max_digits=4, decimal_places=2, db_index=True, null=True, blank=True
    )  # Tax=None is when tax is not applicable
    currency = models.CharField(_("currency"), max_length=3)
    renewal_triggered_by = models.IntegerField(
        _("renewal triggered by"),
        choices=RENEWAL_TRIGGERED_BY,
        help_text=_(
            "The source of the associated plan's renewal (USER = user-initiated renewal, "
            "TASK = autorenew_account-task-initiated renewal, OTHER = renewal is triggered using another mechanism)."
        ),
        default=RENEWAL_TRIGGERED_BY.USER,
        db_index=True,
    )
    # A backup of the old has_automatic_renewal field to support data migration to the new renewal_triggered_by field.
    # Do not make any other modifications to the field in order to let user's auto-migrations detect the renaming.
    # TODO: _has_automatic_renewal_backup_deprecated deprecated. Remove in the next major release.
    _has_automatic_renewal_backup_deprecated = models.BooleanField(
        _("has automatic plan renewal"),
        help_text=_(
            "Automatic renewal is enabled for associated plan. "
            "If False, the plan renewal can be still initiated by user.",
        ),
        db_column="has_automatic_renewal",
        default=False,
    )
    token_verified = models.BooleanField(
        _("token has been verified by payment"),
        help_text=_(
            "The recurring token has been verified by at least one payment to be working.",
        ),
        default=False,
    )
    card_expire_year = models.IntegerField(null=True, blank=True)
    card_expire_month = models.IntegerField(null=True, blank=True)
    card_masked_number = models.CharField(null=True, blank=True, max_length=255)
    last_renewal_attempt = models.DateTimeField(
        _("last renewal attempt"), null=True, blank=True
    )

    class Meta:
        abstract = True

    # TODO: has_automatic_renewal deprecated. Remove in the next major release.
    @property
    def has_automatic_renewal(self):
        warnings.warn(
            "has_automatic_renewal is deprecated. Use renewal_triggered_by instead.",
            DeprecationWarning,
        )
        return self.renewal_triggered_by != self.RENEWAL_TRIGGERED_BY.USER

    # TODO: has_automatic_renewal deprecated. Remove in the next major release.
    @has_automatic_renewal.setter
    def has_automatic_renewal(self, value):
        warnings.warn(
            "has_automatic_renewal is deprecated. Use renewal_triggered_by instead.",
            DeprecationWarning,
        )
        self.renewal_triggered_by = (
            self.RENEWAL_TRIGGERED_BY.TASK if value else self.RENEWAL_TRIGGERED_BY.USER
        )

    # TODO: has_automatic_renewal deprecated. Remove in the next major release.
    @has_automatic_renewal.deleter
    def has_automatic_renewal(self):
        warnings.warn(
            "has_automatic_renewal is deprecated. Use renewal_triggered_by instead.",
            DeprecationWarning,
        )
        del self.renewal_triggered_by

    def create_renew_order(self):
        """
        Create order for plan renewal
        """
        userplan = self.user_plan
        order = AbstractOrder.get_concrete_model().objects.create(
            user=userplan.user,
            plan=userplan.plan,
            pricing=userplan.recurring.pricing,
            amount=userplan.recurring.amount,
            tax=userplan.recurring.tax,  # Fallback value in case of VIES fault
            currency=userplan.recurring.currency,
        )
        order.recalculate(
            userplan.recurring.amount, userplan.user.billinginfo, use_default=False
        )
        order.save()

        # Save new value of tax
        userplan.recurring.tax = order.tax
        userplan.recurring.save()
        return order

    def set_all_fields_default(self):
        """
        Set all fields to default values
        """
        self.token = None
        self.payment_provider = None
        self.pricing = None
        self.amount = None
        self.tax = None
        self.currency = None
        self.renewal_triggered_by = self.RENEWAL_TRIGGERED_BY.USER
        self.token_verified = False
        self.card_expire_year = None
        self.card_expire_month = None
        self.card_masked_number = None


class AbstractPricing(BaseMixin, models.Model):
    """
    Type of plan period that could be purchased (e.g. 10 days, month, year, etc)
    """

    name = models.CharField(_("name"), max_length=100)
    period = models.PositiveIntegerField(
        _("period"), default=30, null=True, blank=True, db_index=True
    )
    url = models.URLField(
        max_length=200,
        blank=True,
        help_text=_(
            "Optional link to page with more information (for clickable pricing table headers)"
        ),
    )

    class Meta:
        abstract = True
        ordering = ("period",)
        verbose_name = _("Pricing")
        verbose_name_plural = _("Pricings")

    def __str__(self):
        return "%s (%d " % (self.name, self.period) + "%s)" % _("days")


class AbstractQuota(BaseMixin, OrderedModel):
    """
    Single countable or boolean property of system (limitation).
    """

    codename = models.CharField(
        _("codename"), max_length=50, unique=True, db_index=True
    )
    name = models.CharField(_("name"), max_length=100)
    unit = models.CharField(_("unit"), max_length=100, blank=True)
    description = models.TextField(_("description"), blank=True)
    is_boolean = models.BooleanField(_("is boolean"), default=False)
    url = models.CharField(
        max_length=200,
        blank=True,
        help_text=_(
            "Optional link to page with more information (for clickable pricing table headers)"
        ),
    )

    class Meta:
        abstract = True
        ordering = ("order",)
        verbose_name = _("Quota")
        verbose_name_plural = _("Quotas")

    def __str__(self):
        return "%s" % (self.codename,)


class PlanPricingManager(models.Manager):
    def get_queryset(self):
        return (
            super(PlanPricingManager, self)
            .get_queryset()
            .select_related("plan", "pricing")
        )


class AbstractPlanPricing(BaseMixin, models.Model):
    plan = models.ForeignKey("Plan", on_delete=models.CASCADE)
    pricing = models.ForeignKey("Pricing", on_delete=models.CASCADE)
    price = models.DecimalField(max_digits=7, decimal_places=2, db_index=True)
    order = models.IntegerField(default=0, null=False, blank=False)
    has_automatic_renewal = models.BooleanField(
        _("has automatic renewal"),
        help_text=_("Use automatic renewal if possible?"),
        default=False,
    )
    visible = models.BooleanField(
        _("is visible by default"),
        help_text=_("Is visible in current offer"),
        default=True,
    )

    objects = PlanPricingManager()

    class Meta:
        abstract = True
        ordering = (
            "order",
            "pricing__period",
        )
        verbose_name = _("Plan pricing")
        verbose_name_plural = _("Plans pricings")

    def __str__(self):
        return f"{self.plan.name} {self.pricing} {'recurring' if self.has_automatic_renewal else ''}"


class PlanQuotaManager(models.Manager):
    def get_queryset(self):
        return (
            super(PlanQuotaManager, self).get_queryset().select_related("plan", "quota")
        )


class AbstractPlanQuota(BaseMixin, models.Model):
    plan = models.ForeignKey("Plan", on_delete=models.CASCADE)
    quota = models.ForeignKey("Quota", on_delete=models.CASCADE)
    value = models.BigIntegerField(default=1, null=True, blank=True)

    objects = PlanQuotaManager()

    class Meta:
        abstract = True
        verbose_name = _("Plan quota")
        verbose_name_plural = _("Plans quotas")


class AbstractOrder(BaseMixin, models.Model):
    """
    Order in this app supports only one item per order. This item is defined by
    plan and pricing attributes. If both are defined the order represents buying
    an account extension.

    If only plan is provided (with pricing set to None) this means that user purchased
    a plan upgrade.
    """

    STATUS = Enumeration(
        [
            (1, "NEW", pgettext_lazy("Order status", "new")),
            (2, "COMPLETED", pgettext_lazy("Order status", "completed")),
            (3, "NOT_VALID", pgettext_lazy("Order status", "not valid")),
            (4, "CANCELED", pgettext_lazy("Order status", "canceled")),
            (5, "RETURNED", pgettext_lazy("Order status", "returned")),
        ]
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_("user"), on_delete=models.CASCADE
    )
    flat_name = models.CharField(max_length=200, blank=True, null=True)
    plan = models.ForeignKey(
        "Plan",
        verbose_name=_("plan"),
        related_name="plan_order",
        on_delete=models.CASCADE,
    )
    pricing = models.ForeignKey(
        "Pricing",
        blank=True,
        null=True,
        verbose_name=_("pricing"),
        on_delete=models.CASCADE,
    )  # if pricing is None the order is upgrade plan, not buy new pricing
    completed = models.DateTimeField(
        _("completed"), null=True, blank=True, db_index=True
    )
    plan_extended_from = models.DateField(
        _("plan extended from"),
        help_text=_("The plan was extended from this date"),
        null=True,
        blank=True,
    )
    plan_extended_until = models.DateField(
        _("plan extended until"),
        help_text=("The plan was extended until this date"),
        null=True,
        blank=True,
    )
    amount = models.DecimalField(
        _("amount"), max_digits=7, decimal_places=2, db_index=True
    )
    tax = models.DecimalField(
        _("tax"), max_digits=4, decimal_places=2, db_index=True, null=True, blank=True
    )  # Tax=None is when tax is not applicable
    currency = models.CharField(_("currency"), max_length=3, default="EUR")
    status = models.IntegerField(_("status"), choices=STATUS, default=STATUS.NEW)

    def __str__(self):
        return _("Order #%(id)d") % {"id": self.id}

    @property
    def name(self):
        """
        Support for two kind of Order names:
        * (preferred) dynamically generated from Plan and Pricing (if flatname is not provided) (translatable)
        * (legacy) just return flat name, which is any text (not translatable)

        Flat names are only introduced for legacy system support, when you need to migrate old orders into
        django-plans and you cannot match Plan&Pricings convention.
        """
        if self.flat_name:
            return self.flat_name
        else:
            return "%s %s %s " % (
                _("Plan"),
                self.plan.name,
                "(upgrade)" if self.pricing is None else "- %s" % self.pricing,
            )

    def is_ready_for_payment(self):
        return self.status == self.STATUS.NEW and (now() - self.created).days < getattr(
            settings, "PLANS_ORDER_EXPIRATION", 14
        )

    def get_plan_extended_from(self):
        return self.user.userplan.get_plan_extended_from(self.plan)

    def get_plan_extended_until(self):
        return self.user.userplan.get_plan_extended_until(self.plan, self.pricing)

    @transaction.atomic()
    def complete_order(self):
        # Get locked order to ensure only one completed order is processed at a time
        order = (
            AbstractOrder.get_concrete_model()
            .objects.filter(id=self.id)
            .select_for_update()
            .get()
        )

        if order.completed is None:
            self.plan_extended_from = self.get_plan_extended_from()
            status = self.user.userplan.extend_account(self.plan, self.pricing)
            self.plan_extended_until = self.user.userplan.expire
            order.completed = self.completed = now()
            if status:
                self.status = self.STATUS.COMPLETED
            else:
                self.status = self.STATUS.NOT_VALID
            self.save()
            order_completed.send(self)
            return True
        else:
            return False

    def return_order(self):
        if self.status != self.STATUS.RETURNED:
            if self.status == self.STATUS.COMPLETED:
                if self.pricing is not None:
                    extended_from = self.plan_extended_from
                    if extended_from is None:
                        extended_from = self.completed
                    # Should never happen, but make sure we reduce for the same number of days as we extended.
                    if (
                        self.plan_extended_until is None
                        or extended_from is None
                        or self.plan_extended_until - extended_from
                        != timedelta(days=self.pricing.period)
                    ):
                        raise ValueError(
                            f"Invalid order state: completed={self.completed}, "
                            f"plan_extended_from={self.plan_extended_from}, "
                            f"plan_extended_until={self.plan_extended_until}, "
                            f"pricing.period={self.pricing.period}"
                        )
                self.user.userplan.reduce_account(self.pricing)
            elif self.status != self.STATUS.NOT_VALID:
                raise ValueError(
                    f"Cannot return order with status other than COMPLETED and NOT_VALID: {self.status}"
                )
            self.status = self.STATUS.RETURNED
            for invoice in self.get_invoices():
                invoice.cancel_invoice()
            self.save()

    def get_invoices_proforma(self):
        return AbstractInvoice.get_concrete_model().proforma.filter(order=self)

    def get_invoices(self):
        return AbstractInvoice.get_concrete_model().invoices.filter(order=self)

    def get_credit_notes(self):
        return AbstractInvoice.get_concrete_model().credit_notes.filter(order=self)

    def get_all_invoices(self):
        return self.invoice_set.order_by("issued", "issued_duplicate", "pk")

    def get_plan_pricing(self):
        return AbstractPlanPricing.get_concrete_model().objects.get(
            plan=self.plan, pricing=self.pricing
        )

    def tax_total(self):
        if self.tax is None:
            return Decimal("0.00")
        else:
            return self.total() - self.amount

    def total(self):
        if self.tax is not None:
            return (Decimal(self.amount) * (Decimal(self.tax) + 100) / 100).quantize(
                Decimal("1.00")
            )
        else:
            return self.amount

    def get_absolute_url(self):
        return reverse("order", kwargs={"pk": self.pk})

    def recalculate(self, amount, billing_info, request=None, use_default=True):
        """
        Calculates and return pre-filled Order
        """
        self.amount = amount
        self.currency = get_currency()

        country = getattr(billing_info, "country", None)
        if country is None:
            country = get_country_code(request)
        else:
            country = country.code

        if hasattr(billing_info, "tax_number") and billing_info.tax_number:
            from plans.base.models import AbstractBillingInfo

            tax_number = AbstractBillingInfo.get_full_tax_number(
                billing_info.tax_number, country
            )
        else:
            tax_number = None

        tax_rate, request_successful = utils.get_tax_rate(country, tax_number, request)
        if (
            use_default or request_successful
        ):  # Don't change the tax, if the request was not successful
            self.tax = tax_rate

    class Meta:
        ordering = ("-created",)
        abstract = True
        verbose_name = _("Order")
        verbose_name_plural = _("Orders")


class InvoiceManager(models.Manager):
    def get_queryset(self):
        return (
            super(InvoiceManager, self)
            .get_queryset()
            .filter(type=AbstractInvoice.INVOICE_TYPES["INVOICE"])
        )


class InvoiceProformaManager(models.Manager):
    def get_queryset(self):
        return (
            super(InvoiceProformaManager, self)
            .get_queryset()
            .filter(type=AbstractInvoice.INVOICE_TYPES["PROFORMA"])
        )


class InvoiceDuplicateManager(models.Manager):
    def get_queryset(self):
        return (
            super(InvoiceDuplicateManager, self)
            .get_queryset()
            .filter(type=AbstractInvoice.INVOICE_TYPES["DUPLICATE"])
        )


class CreditNoteManager(models.Manager):
    def get_queryset(self):
        return (
            super(CreditNoteManager, self)
            .get_queryset()
            .filter(type=AbstractInvoice.INVOICE_TYPES["CREDIT_NOTE"])
        )


def get_initial_number(older_invoices):
    return (
        older_invoices.order_by("number").values_list("number", flat=True).last() or 0
    ) + 1


class AbstractInvoice(BaseMixin, models.Model):
    """
    Single invoice document.
    """

    INVOICE_TYPES = Enumeration(
        [
            (1, "INVOICE", _("Invoice")),
            (2, "DUPLICATE", _("Invoice Duplicate")),
            (3, "PROFORMA", pgettext_lazy("proforma", "Order confirmation")),
            (4, "CREDIT_NOTE", _("Credit note")),
        ]
    )

    objects = models.Manager()
    invoices = InvoiceManager()
    proforma = InvoiceProformaManager()
    duplicates = InvoiceDuplicateManager()
    credit_notes = CreditNoteManager()

    class NUMBERING:
        """Used as a choices for settings.PLANS_INVOICE_COUNTER_RESET"""

        DAILY = 1
        MONTHLY = 2
        ANNUALLY = 3

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_("user"), on_delete=models.CASCADE
    )
    order = models.ForeignKey(
        "Order", verbose_name=_("order"), on_delete=models.CASCADE
    )
    credit_note_for = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="issued_credit_notes",
        verbose_name=_("Credit note for"),
    )
    number = models.IntegerField(db_index=True)
    full_number = models.CharField(max_length=200)
    type = models.IntegerField(
        choices=INVOICE_TYPES, default=INVOICE_TYPES.INVOICE, db_index=True
    )
    issued = models.DateField(db_index=True)
    issued_duplicate = models.DateField(db_index=True, null=True, blank=True)
    selling_date = models.DateField(db_index=True, null=True, blank=True)
    payment_date = models.DateField(db_index=True)
    unit_price_net = models.DecimalField(max_digits=7, decimal_places=2)
    quantity = models.IntegerField(default=1)
    total_net = models.DecimalField(max_digits=7, decimal_places=2)
    total = models.DecimalField(max_digits=7, decimal_places=2)
    tax_total = models.DecimalField(max_digits=7, decimal_places=2)
    tax = models.DecimalField(
        max_digits=4, decimal_places=2, db_index=True, null=True, blank=True
    )  # Tax=None is whet tax is not applicable
    rebate = models.DecimalField(max_digits=4, decimal_places=2, default=Decimal(0))
    currency = models.CharField(max_length=3, default="EUR")
    item_description = models.CharField(max_length=200)
    buyer_name = models.CharField(max_length=200, verbose_name=_("Name"), blank=True)
    buyer_street = models.CharField(
        max_length=200, verbose_name=_("Street"), blank=True
    )
    buyer_zipcode = models.CharField(
        max_length=200, verbose_name=_("Zip code"), blank=True
    )
    buyer_city = models.CharField(max_length=200, verbose_name=_("City"), blank=True)
    buyer_country = CountryField(verbose_name=_("Country"), default="PL", blank=True)
    buyer_tax_number = models.CharField(
        max_length=200, blank=True, verbose_name=_("TAX/VAT number")
    )
    shipping_name = models.CharField(max_length=200, verbose_name=_("Name"), blank=True)
    shipping_street = models.CharField(
        max_length=200, verbose_name=_("Street"), blank=True
    )
    shipping_zipcode = models.CharField(
        max_length=200, verbose_name=_("Zip code"), blank=True
    )
    shipping_city = models.CharField(max_length=200, verbose_name=_("City"), blank=True)
    shipping_country = CountryField(verbose_name=_("Country"), default="PL", blank=True)
    require_shipment = models.BooleanField(default=False, db_index=True)
    issuer_name = models.CharField(max_length=200, verbose_name=_("Name"))
    issuer_street = models.CharField(max_length=200, verbose_name=_("Street"))
    issuer_zipcode = models.CharField(max_length=200, verbose_name=_("Zip code"))
    issuer_city = models.CharField(max_length=200, verbose_name=_("City"))
    issuer_country = CountryField(verbose_name=_("Country"), default="PL")
    issuer_tax_number = models.CharField(
        max_length=200, blank=True, verbose_name=_("TAX/VAT number")
    )
    cancellation_reason = models.TextField(_("Cancellation reason"), blank=True)

    class Meta:
        abstract = True
        verbose_name = _("Invoice")
        verbose_name_plural = _("Invoices")

    def __str__(self):
        return self.full_number

    def get_absolute_url(self):
        return reverse("invoice_preview_html", kwargs={"pk": self.pk})

    def clean(self):
        if self.number is None:
            Invoice = self.get_concrete_model()
            invoice_counter_reset = getattr(
                settings, "PLANS_INVOICE_COUNTER_RESET", Invoice.NUMBERING.MONTHLY
            )
            invoice_counter_reset_name = invoice_counter_reset

            # To avoid duplicates as well as gaps in the sequence, we are using django-sequences
            # to generate sequence number for each invoice
            # We keep the old sequence generating mechanism to get lower initial value,
            # so that the sequence will continue backward compatibly
            older_invoices = Invoice.objects.filter(type=self.type)
            initial_number = None
            if invoice_counter_reset == Invoice.NUMBERING.DAILY:
                invoice_counter_value = (
                    f"{self.issued.year}_{self.issued.month}_{self.issued.day}"
                )
                older_invoices = older_invoices.filter(issued=self.issued)
            elif invoice_counter_reset == Invoice.NUMBERING.MONTHLY:
                invoice_counter_value = f"{self.issued.year}_{self.issued.month}"
                older_invoices = older_invoices.filter(
                    issued__year=self.issued.year,
                    issued__month=self.issued.month,
                )
            elif invoice_counter_reset == Invoice.NUMBERING.ANNUALLY:
                invoice_counter_value = f"{self.issued.year}"
                older_invoices = older_invoices.filter(issued__year=self.issued.year)
            elif callable(invoice_counter_reset):
                invoice_counter_value, initial_number = invoice_counter_reset(self)
                invoice_counter_reset_name = "call"
            else:
                raise ImproperlyConfigured(
                    "PLANS_INVOICE_COUNTER_RESET can be set only to these values: daily, monthly, yearly."
                )

            # get initial value for backward compatibility
            if initial_number:
                self.initial_number = initial_number
            else:
                self.initial_number = get_initial_number(older_invoices)
            self.sequence_name = f"invoice_numbers_{self.type}_{invoice_counter_reset_name}_{invoice_counter_value}"

    def save(self, *args, **kwargs):
        with transaction.atomic():
            if self.number is None:
                self.number = get_next_value(
                    self.sequence_name, initial_value=self.initial_number
                )
            super(AbstractInvoice, self).save(*args, **kwargs)

        # We need to generate full number based on what invoice sequence number actually ended up in DB
        self.refresh_from_db()
        if self.full_number == "":
            self.full_number = self.get_full_number()
        super(AbstractInvoice, self).save(update_fields=["full_number"])

    #    def validate_unique(self, exclude=None):
    #        super(Invoice, self).validate_unique(exclude)
    #        if self.type == Invoice.INVOICE_TYPES.INVOICE:
    #            if Invoice.objects.filter(order=self.order).count():
    #                raise ValidationError("Duplicate invoice for order")
    #        if self.type in (Invoice.INVOICE_TYPES.INVOICE, Invoice.INVOICE_TYPES.PROFORMA):
    #            pass

    def get_full_number(self):
        """
        Generates on the fly invoice full number from template provided by ``settings.PLANS_INVOICE_NUMBER_FORMAT``.
        ``Invoice`` object is provided as ``invoice`` variable to the template, therefore all object fields
        can be used to generate full number format.

        .. warning::

            This is only used to prepopulate ``full_number`` field on saving new invoice.
            To get invoice full number always use ``full_number`` field.

        :return: string (generated full number)
        """
        format = getattr(
            settings,
            "PLANS_INVOICE_NUMBER_FORMAT",
            "{{ invoice.number }}/"
            "{% if invoice.type == invoice.INVOICE_TYPES.PROFORMA %}PF{% else %}FV{% endif %}"
            "/{{ invoice.issued|date:'m/Y' }}",
        )
        return Template(format).render(Context({"invoice": self}))

    def set_issuer_invoice_data(self):
        """
        Fills models object with issuer data copied from ``settings.PLANS_INVOICE_ISSUER``

        :raise: ImproperlyConfigured
        """
        try:
            issuer = getattr(settings, "PLANS_INVOICE_ISSUER")
        except Exception:
            raise ImproperlyConfigured(
                "Please set PLANS_INVOICE_ISSUER in order to make an invoice."
            )
        self.issuer_name = issuer["issuer_name"]
        self.issuer_street = issuer["issuer_street"]
        self.issuer_zipcode = issuer["issuer_zipcode"]
        self.issuer_city = issuer["issuer_city"]
        self.issuer_country = issuer["issuer_country"]
        self.issuer_tax_number = issuer["issuer_tax_number"]

    def set_buyer_invoice_data(self, billing_info):
        """
        Fill buyer invoice billing and shipping data by copy them from provided user's ``BillingInfo`` object.

        :param billing_info: BillingInfo object
        :type billing_info: BillingInfo
        """
        self.buyer_name = billing_info.name
        self.buyer_street = billing_info.street
        self.buyer_zipcode = billing_info.zipcode
        self.buyer_city = billing_info.city
        self.buyer_country = billing_info.country
        self.buyer_tax_number = billing_info.tax_number

        self.shipping_name = billing_info.shipping_name or billing_info.name
        self.shipping_street = billing_info.shipping_street or billing_info.street
        self.shipping_zipcode = billing_info.shipping_zipcode or billing_info.zipcode
        self.shipping_city = billing_info.shipping_city or billing_info.city
        # TODO: Should allow shipping to other country? Not think so
        self.shipping_country = billing_info.country

    def copy_from_order(self, order):
        """
        Filling orders details likes totals, taxes, etc and linking provided ``Order`` object with an invoice

        :param order: Order object
        :type order: Order
        """
        self.order = order
        self.user = order.user
        self.unit_price_net = order.amount
        self.total_net = order.amount
        self.total = order.total()
        self.tax_total = order.total() - order.amount
        self.tax = order.tax
        self.currency = order.currency
        if Site is not None:
            self.item_description = "%s - %s" % (
                Site.objects.get_current().name,
                order.name,
            )
        else:
            self.item_description = order.name

    @classmethod
    def create(cls, order, invoice_type):
        language_code = get_user_language(order.user)

        if language_code is not None:
            translation.activate(language_code)

        BillingInfo = AbstractBillingInfo.get_concrete_model()
        try:
            billing_info = BillingInfo.objects.get(user=order.user)
        except BillingInfo.DoesNotExist:
            return

        day = date.today()
        pday = order.completed
        if invoice_type == cls.INVOICE_TYPES["PROFORMA"]:
            pday = day + timedelta(days=14)

        invoice = cls(
            issued=day, selling_date=order.completed, payment_date=pday
        )  # FIXME: 14 - this should set accordingly to ORDER_TIMEOUT in days
        invoice.type = invoice_type
        invoice.copy_from_order(order)
        invoice.set_issuer_invoice_data()
        invoice.set_buyer_invoice_data(billing_info)
        invoice.clean()
        invoice.save()
        if language_code is not None:
            translation.deactivate()

        return invoice

    def send_invoice_by_email(self):
        if self.type in getattr(
            settings, "PLANS_SEND_EMAILS_DISABLED_INVOICE_TYPES", []
        ):
            return

        language_code = get_user_language(self.user)

        if language_code is not None:
            translation.activate(language_code)
        mail_context = {
            "user": self.user,
            "invoice_type": self.get_type_display(),
            "invoice_number": self.get_full_number(),
            "order": self.order.id,
            "order_object": self.order,
            "url": self.get_absolute_url(),
        }
        if language_code is not None:
            translation.deactivate()
        send_template_email(
            [self.user.email],
            "mail/invoice_created_title.txt",
            "mail/invoice_created_body.txt",
            mail_context,
            language_code,
        )

    def is_UE_customer(self):
        return EUTaxationPolicy.is_in_EU(self.buyer_country.code)

    def cancel_invoice(self, reason=""):
        """
        Cancels the invoice by creating a credit note.
        It also returns the associated order.
        """
        Invoice = self.get_concrete_model()
        if self.type != Invoice.INVOICE_TYPES.INVOICE:
            raise ValidationError(_("Only invoices can be cancelled."))

        if self.issued_credit_notes.exists():
            raise ValidationError(
                _("This invoice has already been cancelled by a credit note.")
            )

        # Create the credit note
        credit_note = Invoice()
        credit_note.type = Invoice.INVOICE_TYPES.CREDIT_NOTE
        credit_note.credit_note_for = self
        credit_note.cancellation_reason = reason

        # Copy data from original invoice
        credit_note.user = self.user
        credit_note.order = self.order
        credit_note.issued = date.today()
        credit_note.payment_date = (
            self.payment_date
        )  # Use original invoice's payment_date
        credit_note.selling_date = self.selling_date

        # Negate values
        # Guard optional fields: subclasses may set unit_price_net, rebate,
        # or item_description to None to indicate they don't use them.
        if self.unit_price_net is not None:
            credit_note.unit_price_net = -self.unit_price_net
        credit_note.quantity = self.quantity
        credit_note.total_net = -self.total_net
        credit_note.total = -self.total
        credit_note.tax_total = -self.tax_total
        credit_note.tax = self.tax  # Tax rate stays the same
        if self.rebate is not None:
            credit_note.rebate = self.rebate
        credit_note.currency = self.currency

        if self.item_description is not None:
            credit_note.item_description = (
                _("Credit note for invoice %s") % self.full_number
            )
        self._copy_addresses_to_credit_note(credit_note)

        # Clean and save
        credit_note.clean()  # This sets up numbering
        credit_note.save()

        return credit_note

    def create_partial_credit_note(self, net_amount, tax_amount, reason=""):
        """
        Creates a partial credit note with custom amounts.
        User input: positive = refund to customer, negative = additional charge.
        """
        Invoice = self.get_concrete_model()
        if self.type != Invoice.INVOICE_TYPES.INVOICE:
            raise ValidationError(_("Only invoices can have partial credit notes."))

        # Build kwargs, skipping fields that subclasses may have set to None
        kwargs = dict(
            type=Invoice.INVOICE_TYPES.CREDIT_NOTE,
            credit_note_for=self,
            user=self.user,
            order=self.order,
            issued=date.today(),
            payment_date=self.payment_date,  # Use original invoice's payment_date
            selling_date=self.selling_date,
            quantity=1,
            total_net=-net_amount,
            tax_total=-tax_amount,
            total=-(net_amount + tax_amount),
            tax=self.tax,
            currency=self.currency,
        )
        if self.unit_price_net is not None:
            kwargs["unit_price_net"] = -net_amount
        if self.rebate is not None:
            kwargs["rebate"] = self.rebate
        credit_note = Invoice(**kwargs)

        total_user_input = net_amount + tax_amount
        action = (
            "Refund"
            if total_user_input > 0
            else "Additional charge" if total_user_input < 0 else "Adjustment"
        )
        if self.item_description is not None:
            credit_note.item_description = _("%s for invoice %s: %s") % (
                action,
                self.full_number,
                reason,
            )
        credit_note.cancellation_reason = "%s: %s" % (action, reason)

        self._copy_addresses_to_credit_note(credit_note)
        credit_note.clean()
        credit_note.save()
        return credit_note

    def _copy_addresses_to_credit_note(self, credit_note):
        """Copy all address fields from invoice to credit note."""
        address_prefixes = ("buyer_", "shipping_", "issuer_")
        address_fields = [
            field.name
            for field in self._meta.fields
            if any(field.name.startswith(prefix) for prefix in address_prefixes)
            or field.name == "require_shipment"
        ]
        for field_name in address_fields:
            setattr(credit_note, field_name, getattr(self, field_name))



"""
Tests for Plan model slug functionality.
"""

from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.test import RequestFactory, TestCase

from plans.admin import PlanAdmin
from plans.base.models import AbstractPlan

User = get_user_model()
Plan = AbstractPlan.get_concrete_model()


class PlanSlugTestCase(TestCase):
    """Test slug functionality for Plan model."""

    def setUp(self):
        """Set up test data."""
        self.factory = RequestFactory()
        # Clear any existing plans to avoid conflicts
        Plan.objects.all().delete()

    def test_slug_field_exists(self):
        """Test that slug field exists and can be set."""
        plan = Plan.objects.create(
            name="Premium Plan",
            slug="premium-plan",
            description="A premium plan for testing",
            available=True,
            visible=True,
        )

        self.assertEqual(plan.slug, "premium-plan")

    def test_slug_can_be_set_manually(self):
        """Test that slug can be set manually."""
        plan = Plan.objects.create(
            name="Pro Plan & More!",
            slug="custom-pro-slug",
            description="Plan with custom slug",
            available=True,
            visible=True,
        )

        self.assertEqual(plan.slug, "custom-pro-slug")

    def test_slug_not_overwritten_if_provided(self):
        """Test that manually provided slug is not overwritten."""
        plan = Plan.objects.create(
            name="Custom Plan",
            slug="my-custom-slug",
            description="Plan with custom slug",
            available=True,
            visible=True,
        )

        self.assertEqual(plan.slug, "my-custom-slug")

    def test_slug_uniqueness_constraint(self):
        """Test that slug field enforces uniqueness at database level."""
        Plan.objects.create(
            name="First Plan",
            slug="test-slug",
            description="First plan",
            available=True,
            visible=True,
        )

        # Second plan with same slug should raise IntegrityError
        plan2 = Plan(
            name="Second Plan",
            slug="test-slug",  # Same slug
            description="Second plan",
            available=True,
            visible=True,
        )

        with self.assertRaises(IntegrityError):
            plan2.save()

    def test_plan_str_method_unchanged(self):
        """Test that __str__ method still returns name."""
        plan = Plan.objects.create(
            name="Test Plan",
            slug="test-plan",
            description="Test plan",
            available=True,
            visible=True,
        )

        self.assertEqual(str(plan), "Test Plan")


class PlanSlugMigrationTestCase(TestCase):
    """Test migration logic for populating slugs from existing names."""

    def test_migration_slug_population_logic(self):
        """Test the logic used in migration to populate slugs."""
        # Test the slugify logic that would be used in migration
        from django.utils.text import slugify

        # Test various name patterns
        test_cases = [
            ("Basic Plan", "basic-plan"),
            ("Premium Plan & More!", "premium-plan-more"),
            ("Plán Básico", "plan-basico"),
            ("Plan   With   Spaces", "plan-with-spaces"),
        ]

        for name, expected_slug in test_cases:
            actual_slug = slugify(name)
            self.assertEqual(actual_slug, expected_slug, f"Failed for name: {name}")

    def test_slug_uniqueness_enforced(self):
        """Test that slug uniqueness is enforced."""
        # Create first plan
        Plan.objects.create(
            name="Standard Plan",
            slug="standard-plan",
            description="First standard plan",
            available=True,
            visible=True,
        )

        # Try to create second plan with same slug - should fail
        with self.assertRaises(IntegrityError):
            Plan.objects.create(
                name="Another Plan",
                slug="standard-plan",  # Same slug
                description="Second plan with same slug",
                available=True,
                visible=True,
            )


class PlanAdminSlugTestCase(TestCase):
    """Test admin functionality for Plan slug field."""

    def setUp(self):
        """Set up admin test data."""
        self.factory = RequestFactory()
        self.site = AdminSite()
        self.admin = PlanAdmin(Plan, self.site)

        # Create a superuser for admin tests
        self.superuser = User.objects.create_superuser(
            username="admin", email="admin@test.com", password="adminpass"
        )

    def test_slug_in_list_display(self):
        """Test that slug is included in admin list display."""
        self.assertIn("slug", self.admin.list_display)

    def test_slug_in_search_fields(self):
        """Test that slug is included in admin search fields."""
        self.assertIn("slug", self.admin.search_fields)

    def test_prepopulated_fields_for_new_objects(self):
        """Test that slug is prepopulated from name for new objects."""
        self.assertEqual(self.admin.prepopulated_fields, {"slug": ("name",)})

    def test_slug_always_editable(self):
        """Test that slug is always editable in admin."""
        # Create a plan
        plan = Plan.objects.create(
            name="Test Plan",
            description="Test plan",
            available=True,
            visible=True,
        )

        # Test readonly fields for existing object
        request = self.factory.get("/admin/plans/plan/")
        request.user = self.superuser

        readonly_fields = self.admin.get_readonly_fields(request, plan)
        self.assertNotIn("slug", readonly_fields)

        # Test readonly fields for new object
        readonly_fields = self.admin.get_readonly_fields(request, None)
        self.assertNotIn("slug", readonly_fields)

    def test_admin_preserves_base_readonly_fields(self):
        """Test that admin preserves base readonly fields."""
        request = self.factory.get("/admin/plans/plan/")
        request.user = self.superuser

        readonly_fields = self.admin.get_readonly_fields(request, None)
        self.assertIn("created", readonly_fields)
        self.assertIn("updated_at", readonly_fields)


class PlanSlugEdgeCasesTestCase(TestCase):
    """Test edge cases for Plan slug functionality."""

    def test_slug_max_length_constraint(self):
        """Test that slug field respects max_length constraint."""
        # Try to create plan with slug that's too long
        long_slug = "a" * 101  # 101 characters - exceeds max_length

        with self.assertRaises(Exception):  # Could be ValidationError or DataError
            Plan.objects.create(
                name="Test Plan",
                slug=long_slug,
                description="Plan with too long slug",
                available=True,
                visible=True,
            )

    def test_slug_with_numbers_and_hyphens(self):
        """Test that slug field accepts numbers and hyphens."""
        plan = Plan.objects.create(
            name="Plan 2024 Version 1.5",
            slug="plan-2024-version-1-5",
            description="Plan with numbers and hyphens",
            available=True,
            visible=True,
        )

        self.assertEqual(plan.slug, "plan-2024-version-1-5")

    def test_slug_field_behavior(self):
        """Test slug field behavior."""
        # Slug field should accept valid slugs
        plan = Plan.objects.create(
            name="Test Plan",
            slug="test-plan",
            description="Plan with valid slug",
            available=True,
            visible=True,
        )

        self.assertEqual(plan.slug, "test-plan")



from modeltranslation.translator import TranslationOptions, translator

from plans.base.models import AbstractPlan, AbstractPricing, AbstractQuota

Plan = AbstractPlan.get_concrete_model()
Pricing = AbstractPricing.get_concrete_model()
Quota = AbstractQuota.get_concrete_model()

# Translations for django-plans


class PlanTranslationOptions(TranslationOptions):
    fields = (
        "name",
        "description",
    )


translator.register(Plan, PlanTranslationOptions)


class PricingTranslationOptions(TranslationOptions):
    fields = ("name",)


translator.register(Pricing, PricingTranslationOptions)


class QuotaTranslationOptions(TranslationOptions):
    fields = ("name", "description", "unit")


translator.register(Quota, QuotaTranslationOptions)



from django.core.management import BaseCommand

from plans import tasks


class Command(BaseCommand):
    help = "Expire accounts and send messages"

    def handle(self, *args, **options):  # pragma: no cover
        tasks.expire_account()
        self.stdout.write("accounts was expired")



# Generated by Django 3.0.9 on 2020-08-03 13:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0007_recurringuserplan_card_masked_number"),
    ]

    operations = [
        migrations.AddField(
            model_name="recurringuserplan",
            name="token_verified",
            field=models.BooleanField(
                default=False,
                help_text="The recurring token has been verified by at least one payment to be working.",
                verbose_name="token has been verified by payment",
            ),
        ),
    ]



from copy import deepcopy

from django.contrib import admin
from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from ordered_model.admin import OrderedModelAdmin

from plans.base.models import (
    AbstractBillingInfo,
    AbstractInvoice,
    AbstractOrder,
    AbstractPlan,
    AbstractPlanPricing,
    AbstractPlanQuota,
    AbstractPricing,
    AbstractQuota,
    AbstractRecurringUserPlan,
    AbstractUserPlan,
)

from .forms import PartialCreditNoteForm
from .signals import account_automatic_renewal

Invoice = AbstractInvoice.get_concrete_model()
UserPlan = AbstractUserPlan.get_concrete_model()
Plan = AbstractPlan.get_concrete_model()
PlanQuota = AbstractPlanQuota.get_concrete_model()
Quota = AbstractQuota.get_concrete_model()
PlanPricing = AbstractPlanPricing.get_concrete_model()
Pricing = AbstractPricing.get_concrete_model()
RecurringUserPlan = AbstractRecurringUserPlan.get_concrete_model()
Order = AbstractOrder.get_concrete_model()
BillingInfo = AbstractBillingInfo.get_concrete_model()


class UserLinkMixin(object):
    def user_link(self, obj):
        user_model = get_user_model()
        app_label = user_model._meta.app_label
        model_name = user_model._meta.model_name
        change_url = reverse(
            "admin:%s_%s_change" % (app_label, model_name), args=(obj.user.id,)
        )
        return format_html('<a href="{}">{}</a>', change_url, obj.user.username)

    user_link.short_description = "User"
    user_link.allow_tags = True


class PlanQuotaInline(admin.TabularInline):
    model = PlanQuota


class PlanPricingInline(admin.TabularInline):
    model = PlanPricing


class QuotaAdmin(OrderedModelAdmin):
    list_display = [
        "codename",
        "name",
        "description",
        "unit",
        "is_boolean",
        "move_up_down_links",
    ]

    readonly_fields = ("created", "updated_at")
    list_display_links = ["codename", "name"]


def copy_plan(modeladmin, request, queryset):
    """
    Admin command for duplicating plans preserving quotas and pricings.
    """
    for plan in queryset:
        plan_copy = deepcopy(plan)
        plan_copy.id = None
        plan_copy.available = False
        plan_copy.default = False
        plan_copy.created = None
        plan_copy.save(force_insert=True)

        for pricing in plan.planpricing_set.all():
            pricing.id = None
            pricing.plan = plan_copy
            pricing.save(force_insert=True)

        for quota in plan.planquota_set.all():
            quota.id = None
            quota.plan = plan_copy
            quota.save(force_insert=True)


copy_plan.short_description = _("Make a plan copy")


class PlanAdmin(OrderedModelAdmin):
    search_fields = (
        "name",
        "slug",
        "customized__username",
        "customized__email",
    )
    list_filter = ("available", "visible")
    list_display = [
        "name",
        "slug",
        "description",
        "customized",
        "default",
        "available",
        "is_free",
        "created",
        "move_up_down_links",
    ]
    list_display_links = ["name", "slug"]
    inlines = (PlanPricingInline, PlanQuotaInline)
    list_select_related = True
    raw_id_fields = ("customized",)
    readonly_fields = ("created", "updated_at")
    prepopulated_fields = {"slug": ("name",)}
    actions = [
        copy_plan,
    ]

    def queryset(self, request):
        return super(PlanAdmin, self).queryset(request).select_related("customized")


class BillingInfoAdmin(UserLinkMixin, admin.ModelAdmin):
    search_fields = ("user__username", "user__email", "tax_number", "name")
    list_display = (
        "user",
        "tax_number",
        "name",
        "street",
        "zipcode",
        "city",
        "country",
    )
    list_display_links = list_display
    list_select_related = True
    readonly_fields = ("user_link", "created", "updated_at")
    exclude = ("user",)


def make_order_completed(modeladmin, request, queryset):
    for order in queryset:
        order.complete_order()


make_order_completed.short_description = _("Make selected orders completed")


def make_order_returned(modeladmin, request, queryset):
    for order in queryset:
        order.return_order()


make_order_returned.short_description = _("Make selected orders returned")


def make_order_invoice(modeladmin, request, queryset):
    for order in queryset:
        if (
            Invoice.objects.filter(
                type=Invoice.INVOICE_TYPES["INVOICE"], order=order
            ).count()
            == 0
        ):
            if not order.completed:
                modeladmin.message_user(
                    request,
                    f"Order {order.id} is has no completed date, cannot create invoice.",
                    level="ERROR",
                )
                continue
            Invoice.create(order, Invoice.INVOICE_TYPES["INVOICE"])


make_order_invoice.short_description = _("Make invoices for orders")


class InvoiceInline(admin.TabularInline):
    model = Invoice
    extra = 0
    raw_id_fields = ("user",)


class OrderAdmin(admin.ModelAdmin):
    list_filter = ("status", "created", "completed", "plan__name", "pricing")
    raw_id_fields = ("user",)
    search_fields = ("id", "user__username", "user__email", "invoice__full_number")
    list_display = (
        "id",
        "name",
        "created",
        "user",
        "status",
        "completed",
        "tax",
        "amount",
        "currency",
        "plan",
        "pricing",
        "plan_extended_from",
        "plan_extended_until",
    )
    readonly_fields = ("created", "updated_at")
    list_display_links = list_display
    actions = [make_order_completed, make_order_returned, make_order_invoice]
    inlines = (InvoiceInline,)

    def queryset(self, request):
        return (
            super(OrderAdmin, self)
            .queryset(request)
            .select_related("plan", "pricing", "user")
        )


def cancel_selected_invoices(modeladmin, request, queryset):
    for invoice in queryset:
        try:
            invoice.cancel_invoice()
            modeladmin.message_user(
                request, f"Invoice {invoice.full_number} cancelled successfully."
            )
        except Exception as e:
            modeladmin.message_user(
                request, f"Could not cancel {invoice.full_number}: {e}", level="ERROR"
            )


cancel_selected_invoices.short_description = _("Cancel and issue credit note")


def create_partial_credit_note(modeladmin, request, queryset):
    """Create partial credit notes with custom amounts"""
    if queryset.count() != 1:
        modeladmin.message_user(
            request,
            "Please select exactly one invoice to create a partial credit note.",
            level="ERROR",
        )
        return

    invoice = queryset.first()
    if invoice.type != invoice.INVOICE_TYPES.INVOICE:
        modeladmin.message_user(
            request,
            "Only regular invoices can have partial credit notes.",
            level="ERROR",
        )
        return

    if request.method == "POST" and "apply" in request.POST:
        form = PartialCreditNoteForm(request.POST)
        if form.is_valid():
            credit_note = invoice.create_partial_credit_note(
                form.cleaned_data["net_amount"],
                form.cleaned_data["tax_amount"],
                form.cleaned_data["reason"],
            )
            modeladmin.message_user(
                request,
                "Partial credit note %s created successfully for invoice %s."
                % (credit_note.full_number, invoice.full_number),
            )
            return HttpResponseRedirect(reverse("admin:plans_invoice_changelist"))
    else:
        # Initialize form with default values from invoice
        initial_data = {
            "net_amount": invoice.total_net,
            "tax_amount": invoice.tax_total,
            "reason": "",
        }
        form = PartialCreditNoteForm(initial=initial_data)

    context = {
        "title": "Create Partial Credit Note for Invoice %s" % invoice.full_number,
        "invoice": invoice,
        "opts": modeladmin.model._meta,
        "has_view_permission": modeladmin.has_view_permission(request),
        "form": form,
    }

    return TemplateResponse(
        request, "admin/plans/invoice/partial_credit_note_form.html", context
    )


create_partial_credit_note.short_description = _("Create partial credit note")


class InvoiceAdmin(admin.ModelAdmin):
    search_fields = ("full_number", "buyer_tax_number", "user__username", "user__email")
    list_filter = (
        "type",
        "issued",
        "tax",
        "currency",
        "buyer_country",
    )
    list_display = (
        "full_number",
        "issued",
        "total_net",
        "currency",
        "user",
        "tax",
        "buyer_name",
        "buyer_city",
        "buyer_tax_number",
    )
    readonly_fields = (
        "created",
        "updated_at",
        "full_number",
        "number",
        "total_net",
        "tax_total",
        "total",
    )
    list_display_links = list_display
    list_select_related = True
    raw_id_fields = ("user", "order", "credit_note_for")
    actions = (cancel_selected_invoices, create_partial_credit_note)
    fieldsets = (
        (
            _("Invoice Details"),
            {
                "fields": (
                    "type",
                    "full_number",
                    "number",
                    "user",
                    "order",
                    "credit_note_for",
                    "cancellation_reason",
                )
            },
        ),
        (
            _("Dates"),
            {
                "fields": (
                    "issued",
                    "selling_date",
                    "payment_date",
                    "issued_duplicate",
                    "created",
                    "updated_at",
                )
            },
        ),
        (
            _("Billing Details"),
            {
                "fields": (
                    "item_description",
                    "quantity",
                    "unit_price_net",
                    "tax",
                    "rebate",
                    "currency",
                    "total_net",
                    "tax_total",
                    "total",
                )
            },
        ),
        (
            _("Buyer Details"),
            {
                "fields": (
                    "buyer_name",
                    "buyer_street",
                    "buyer_zipcode",
                    "buyer_city",
                    "buyer_country",
                    "buyer_tax_number",
                )
            },
        ),
        (
            _("Shipping Details"),
            {
                "fields": (
                    "require_shipment",
                    "shipping_name",
                    "shipping_street",
                    "shipping_zipcode",
                    "shipping_city",
                    "shipping_country",
                )
            },
        ),
        (
            _("Issuer Details"),
            {
                "fields": (
                    "issuer_name",
                    "issuer_street",
                    "issuer_zipcode",
                    "issuer_city",
                    "issuer_country",
                    "issuer_tax_number",
                )
            },
        ),
    )


class RecurringPlanInline(admin.StackedInline):
    model = RecurringUserPlan
    readonly_fields = ("created", "updated_at")
    extra = 0


def autorenew_payment(modeladmin, request, queryset):
    """
    Automatically renew payment for this plan
    """
    for user_plan in queryset:
        account_automatic_renewal.send(sender=None, user=user_plan.user)


autorenew_payment.short_description = _("Autorenew plan")


class UserPlanAdmin(UserLinkMixin, admin.ModelAdmin):
    list_filter = (
        "active",
        "expire",
        "plan__name",
        "plan__available",
        "plan__visible",
        "recurring__renewal_triggered_by",
        "recurring__payment_provider",
        "recurring__token_verified",
        "recurring__pricing",
    )
    search_fields = ("user__username", "user__email", "plan__name", "recurring__token")
    list_display = (
        "user",
        "plan",
        "expire",
        "active",
        "recurring__renewal_triggered_by",
        "recurring__token_verified",
        "recurring__payment_provider",
        "recurring__pricing",
    )
    list_display_links = list_display
    list_select_related = True
    readonly_fields = ("user_link", "created", "updated_at")
    inlines = (RecurringPlanInline,)
    actions = [
        autorenew_payment,
    ]
    fields = ("user", "user_link", "plan", "expire", "active", "created", "updated_at")
    raw_id_fields = [
        "user",
        "plan",
    ]

    def recurring__renewal_triggered_by(self, obj):
        return obj.recurring.renewal_triggered_by

    recurring__renewal_triggered_by.admin_order_field = (
        "recurring__renewal_triggered_by"
    )
    recurring__renewal_triggered_by.short_description = "Renewal triggered by"

    def recurring__token_verified(self, obj):
        return obj.recurring.token_verified

    recurring__token_verified.admin_order_field = "recurring__token_verified"
    recurring__token_verified.boolean = True
    recurring__token_verified.short_description = "Renewal token verified"

    def recurring__payment_provider(self, obj):
        return obj.recurring.payment_provider

    recurring__payment_provider.admin_order_field = "recurring__payment_provider"
    recurring__payment_provider.short_description = "Renewal payment_provider"

    def recurring__pricing(self, obj):
        return obj.recurring.pricing

    recurring__pricing.admin_order_field = "recurring__pricing"


admin.site.register(Quota, QuotaAdmin)
admin.site.register(Plan, PlanAdmin)
admin.site.register(UserPlan, UserPlanAdmin)
admin.site.register(Pricing)
admin.site.register(Order, OrderAdmin)
admin.site.register(BillingInfo, BillingInfoAdmin)
admin.site.register(Invoice, InvoiceAdmin)



import logging
from decimal import Decimal
from urllib.error import URLError
from xml.sax import SAXParseException

import stdnum.eu.vat
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.utils.html import format_html
from requests.exceptions import ConnectionError, Timeout
from zeep.exceptions import Fault, TransportError, XMLSyntaxError

from plans.taxation import TaxationPolicy
from plans.taxation.tedb_client import TEDBClient
from plans.utils import country_code_transform

logger = logging.getLogger("plans.taxation.eu.vies")


class EUTaxationPolicy(TaxationPolicy):
    """
    This taxation policy should be correct for all EU countries. It uses following rules:
        * if issuer country is not in EU - assert error,
        * for buyer of the same country as issuer - return issuer tax,
        * for company buyer from EU (with VIES) returns VAT n/a reverse charge,
        * for non-company buyer from EU returns VAT from buyer country,
        * for non-EU buyer return VAT n/a.

    This taxation policy was updated at 1 Jan 2015 after new UE VAT regulations. You should also probably
    register in MOSS system.

    VAT rates are retrieved from the European Commission's TEDB service when possible,
    with fallback to static rates table for reliability.
    """

    # Standard VAT rates for EU countries
    # Updated: November 2025
    # Note: VAT rates change frequently - verify current rates for production use

    EU_COUNTRIES_VAT = {
        "BE": Decimal("21"),  # Belgium
        "BG": Decimal("20"),  # Bulgaria
        "CZ": Decimal("21"),  # Czech Republic
        "DK": Decimal("25"),  # Denmark
        "DE": Decimal("19"),  # Germany
        "EE": Decimal("24"),  # Estonia (increased to 24% in July 2025)
        "EL": Decimal("24"),  # Greece
        "ES": Decimal("21"),  # Spain
        "FR": Decimal("20"),  # France
        "HR": Decimal("25"),  # Croatia
        "IE": Decimal("23"),  # Ireland
        "IT": Decimal("22"),  # Italy
        "CY": Decimal("19"),  # Cyprus
        "LV": Decimal("21"),  # Latvia
        "LT": Decimal("21"),  # Lithuania
        "LU": Decimal("17"),  # Luxembourg
        "HU": Decimal("27"),  # Hungary
        "MT": Decimal("18"),  # Malta
        "NL": Decimal("21"),  # Netherlands
        "AT": Decimal("20"),  # Austria
        "PL": Decimal("23"),  # Poland
        "PT": Decimal("23"),  # Portugal
        "RO": Decimal("21"),  # Romania (increased from 19% in Aug 2025)
        "SI": Decimal("22"),  # Slovenia
        "SK": Decimal("23"),  # Slovakia (increased from 20% in Jan 2025)
        "FI": Decimal("25.5"),  # Finland (increased from 24% in Sep 2024)
        "SE": Decimal("25"),  # Sweden
        # 'GB': Decimal('20'),  # United Kingdom (Great Britain)
    }

    @classmethod
    def is_in_EU(cls, country_code):
        return country_code_transform(country_code).upper() in cls.EU_COUNTRIES_VAT

    @classmethod
    def _get_tedb_client(cls):
        """Get TEDB client instance, cached as class attribute."""
        if not hasattr(cls, "_tedb_client"):
            cls._tedb_client = TEDBClient()
        return cls._tedb_client

    @classmethod
    def _get_vat_rate_from_tedb(cls, country_code):
        """
        Get VAT rate from TEDB service with fallback to static table.

        Args:
            country_code: ISO 2-letter country code

        Returns:
            Decimal VAT rate or None if not available
        """
        try:
            tedb_client = cls._get_tedb_client()
            rate = tedb_client.get_vat_rate(country_code)
            if rate is not None:
                logger.info(f"Using TEDB VAT rate for {country_code}: {rate}")
                return rate
        except (ConnectionError, Timeout, Fault, TransportError, XMLSyntaxError) as e:
            logger.warning(f"TEDB service unavailable for {country_code}: {e}")

        # Fallback to static table
        try:
            rate = cls.EU_COUNTRIES_VAT[country_code]
            logger.info(f"Using fallback VAT rate for {country_code}: {rate}")
            return rate
        except KeyError:
            logger.error(f"No VAT rate available for {country_code}")
            return None

    @classmethod
    def get_default_tax(cls):
        issuer_country_code = cls.get_issuer_country_code()
        issuer_country_code = country_code_transform(issuer_country_code)

        # Try TEDB first, then fallback to static table
        rate = cls._get_vat_rate_from_tedb(issuer_country_code)
        if rate is not None:
            return rate

        raise ImproperlyConfigured(
            "EUTaxationPolicy requires that issuer country is in EU"
        )

    @classmethod
    def get_tax_rate(cls, tax_id, country_code, request=None):
        """
        returns tax rate and if the request was successful.
        """
        country_code = country_code_transform(country_code)
        issuer_country_code = cls.get_issuer_country_code()
        if not cls.is_in_EU(issuer_country_code):
            raise ImproperlyConfigured(
                "EUTaxationPolicy requires that issuer country is in EU"
            )

        if not tax_id and not country_code:
            # No vat id, no country
            return cls.get_default_tax(), True

        elif not tax_id and country_code:
            # Customer is not a company, we know his country

            if cls.is_in_EU(country_code):
                # Customer (private person) is from a EU
                # Customer pays his VAT rate
                rate = cls._get_vat_rate_from_tedb(country_code)
                if rate is not None:
                    return rate, True
                return cls.EU_COUNTRIES_VAT[country_code], True
            else:
                # Customer (private person) not from EU
                # VAT n/a
                return None, True

        else:
            # Customer is company, we now country and vat id

            if country_code.upper() == issuer_country_code.upper():
                # Company is from the same country as issuer
                # Normal tax
                return cls.get_default_tax(), True

            if cls.is_in_EU(country_code):
                # Company is from other EU country
                try:
                    vies_result = bool(stdnum.eu.vat.check_vies(tax_id)["valid"])
                    logger.info("TAX_ID=%s RESULT=%s" % (tax_id, vies_result))
                    if tax_id and vies_result:
                        # Company is registered in VIES
                        # Charge back
                        return None, True
                    else:
                        rate = cls._get_vat_rate_from_tedb(country_code)
                        if rate is not None:
                            return rate, True
                        return cls.EU_COUNTRIES_VAT[country_code], True
                except (
                    Fault,
                    TransportError,
                    XMLSyntaxError,
                    stdnum.exceptions.InvalidComponent,
                    ConnectionError,
                    URLError,
                    SAXParseException,
                    TimeoutError,
                ) as e:
                    # If we could not connect to VIES or the VAT ID is incorrect
                    if request:
                        messages.warning(
                            request,
                            format_html(
                                "There was an error during determining validity of your VAT ID.<br/>"
                                "If you think, you have european VAT ID and should not be taxed, "
                                "please try resaving billing info later.<br/>"
                                "<br/>"
                                "European VAT Information Exchange System throw following error: {}",
                                e,
                            ),
                        )
                    logger.exception("TAX_ID=%s" % (tax_id))
                    rate = cls._get_vat_rate_from_tedb(country_code)
                    if rate is not None:
                        return rate, False
                    return cls.EU_COUNTRIES_VAT[country_code], False
            else:
                # Company is not from EU
                # VAT n/a
                return None, True



# Generated by Django 4.2.11 on 2024-04-09 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("plans", "0012_planpricing_visible"),
    ]

    operations = [
        migrations.AlterField(
            model_name="recurringuserplan",
            name="has_automatic_renewal",
            field=models.BooleanField(
                db_column="has_automatic_renewal",
                default=False,
                help_text="Automatic renewal is enabled for associated plan. If False, the plan renewal can be still initiated by user.",
                verbose_name="has automatic plan renewal",
            ),
        ),
        migrations.RenameField(
            model_name="recurringuserplan",
            old_name="has_automatic_renewal",
            new_name="_has_automatic_renewal_backup_deprecated",
        ),
        migrations.AddField(
            model_name="recurringuserplan",
            name="renewal_triggered_by",
            field=models.IntegerField(
                choices=[(1, "other"), (2, "user"), (3, "task")],
                db_index=True,
                default=2,
                help_text="The source of the associated plan's renewal (USER = user-initiated renewal, TASK = autorenew_account-task-initiated renewal, OTHER = renewal is triggered using another mechanism).",
                verbose_name="renewal triggered by",
            ),
        ),
    ]



"""
Test for migration issue #192 - Error when migrate
https://github.com/django-getpaid/django-plans/issues/192
"""

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

User = get_user_model()


class MigrationIssue192TestCase(TestCase):
    """Test that reproduces and verifies the fix for migration issue.

    Tests the case where updated_at field is referenced before it exists.
    """

    def test_migration_0004_logic_with_historical_models(self):
        """
        Test the fixed migration logic using apps.get_model() instead of importing current models.
        This simulates what the fixed migration 0004_create_user_plans does.
        """
        # Simulate using historical models (like in the migration)
        User = apps.get_model(settings.AUTH_USER_MODEL)
        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Create a user without a user plan
        user = User.objects.create_user(
            username="testuser", email="test@example.com", password="testpass"
        )

        # Create a default plan
        plan = Plan.objects.create(
            name="Default Plan",
            description="Default plan for testing",
            default=True,
            available=True,
            visible=True,
        )

        # Delete any existing user plans to simulate fresh migration
        UserPlan.objects.filter(user=user).delete()

        # This is the fixed migration logic
        try:
            default_plan = Plan.objects.filter(default=True).first()
            self.assertIsNotNone(default_plan)

            users_without_plans = User.objects.filter(userplan=None)
            self.assertGreaterEqual(users_without_plans.count(), 1)

            # Create user plans for users without them
            for user_without_plan in users_without_plans:
                UserPlan.objects.create(
                    user=user_without_plan,
                    plan=default_plan,
                    active=False,
                    expire=None,
                )

            # Verify user plan was created
            user_plan = UserPlan.objects.get(user=user)
            self.assertEqual(user_plan.plan, plan)
            self.assertFalse(user_plan.active)  # Should be inactive by default
            self.assertIsNone(user_plan.expire)

        except Exception as e:
            self.fail(f"Fixed migration logic failed: {e}")

    def test_migration_handles_no_default_plan(self):
        """
        Test that the migration gracefully handles the case where no default plan exists.
        """
        # Simulate using historical models
        User = apps.get_model(settings.AUTH_USER_MODEL)
        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Create a user
        user = User.objects.create_user(
            username="testuser2", email="test2@example.com", password="testpass"
        )

        # Ensure no default plan exists
        Plan.objects.filter(default=True).delete()

        # Delete any existing user plans
        UserPlan.objects.filter(user=user).delete()

        # This should not fail even when no default plan exists
        try:
            default_plan = Plan.objects.filter(default=True).first()
            self.assertIsNone(default_plan)

            if default_plan is None:
                # Should skip creating user plans
                initial_count = UserPlan.objects.filter(user=user).count()
                # Migration logic would return early here
                final_count = UserPlan.objects.filter(user=user).count()
                self.assertEqual(initial_count, final_count)

        except Exception as e:
            self.fail(f"Migration should handle missing default plan gracefully: {e}")



"""
Tests for the slug migration functionality.
"""

from django.test import TestCase
from django.utils.text import slugify

from plans.base.models import AbstractPlan

Plan = AbstractPlan.get_concrete_model()


class SlugMigrationTestCase(TestCase):
    """Test the migration logic for adding and populating slug field."""

    def setUp(self):
        """Set up test data."""
        # Clear any existing plans to avoid conflicts
        Plan.objects.all().delete()

    def test_populate_slugs_migration_function(self):
        """Test the populate_slugs function from migration 0017."""
        # Test the migration logic directly

        # Test cases that would be handled by the migration
        test_cases = [
            ("Basic Plan", "basic-plan"),
            ("Premium Plan & More", "premium-plan-more"),
            ("Enterprise Solution", "enterprise-solution"),
            ("Plan With Duplicate Name", "plan-with-duplicate-name"),
            (
                "Plan With Duplicate Name",
                "plan-with-duplicate-name-1",
            ),  # Would get -1 suffix
        ]

        used_slugs = set()
        for name, expected_base in test_cases:
            base_slug = slugify(name)
            slug = base_slug
            counter = 1

            # Handle duplicates by appending a number (migration logic)
            while slug in used_slugs:
                slug = f"{base_slug}-{counter}"
                counter += 1

            used_slugs.add(slug)

            # Verify the logic works as expected
            if (
                name == "Plan With Duplicate Name"
                and len(
                    [s for s in used_slugs if s.startswith("plan-with-duplicate-name")]
                )
                == 1
            ):
                self.assertEqual(slug, "plan-with-duplicate-name")
            elif name == "Plan With Duplicate Name":
                self.assertEqual(slug, "plan-with-duplicate-name-1")

    def test_migration_with_existing_slugs(self):
        """Test migration behavior when some plans already have slugs."""
        # Create a plan with existing slug
        plan_with_slug = Plan.objects.create(
            name="Custom Plan",
            slug="my-custom-slug",
            description="Plan with existing slug",
            available=True,
            visible=True,
        )

        # Simulate migration by manually clearing and repopulating
        original_slug = plan_with_slug.slug

        # The migration should not overwrite existing slugs
        # (our save method only sets slug if it's empty)
        plan_with_slug.save()

        self.assertEqual(plan_with_slug.slug, original_slug)

    def test_migration_handles_unicode_names(self):
        """Test migration with unicode characters in names."""
        plan = Plan.objects.create(
            name="Plán Básico",
            slug="plan-basico",  # Provide the slug explicitly
            description="Plan with unicode characters",
            available=True,
            visible=True,
        )

        # Should store the provided slug
        self.assertEqual(plan.slug, "plan-basico")

    def test_migration_handles_duplicate_names_from_fixtures(self):
        """Test migration properly handles duplicate plan names like in initial_plan.json."""
        # Simulate the exact scenario from initial_plan.json where there are
        # two plans both named "Default Plan"

        # We can't actually create duplicates due to our validation, so we'll
        # test the migration logic directly

        # Simulate plans with same names (like in fixtures)
        plan_names = ["Default Plan", "Default Plan", "Standard", "Premium"]

        # Test the migration logic
        used_slugs = set()
        generated_slugs = []

        for name in plan_names:
            base_slug = slugify(name)
            slug = base_slug
            counter = 1

            # Handle duplicates by appending a number
            while slug in used_slugs:
                slug = f"{base_slug}-{counter}"
                counter += 1

            used_slugs.add(slug)
            generated_slugs.append(slug)

        # Verify the expected slug generation
        expected_slugs = ["default-plan", "default-plan-1", "standard", "premium"]
        self.assertEqual(generated_slugs, expected_slugs)

    def test_migration_with_long_names(self):
        """Test migration with names that create long slugs."""
        # Use a name that fits in name field but creates a long slug
        long_name = (
            "Plan With Many Words That Create A Very Long Slug"  # Fits in name field
        )
        plan = Plan.objects.create(
            name=long_name,
            slug="plan-with-many-words-that-create-a-very-long-slug",
            description="Plan with long name",
            available=True,
            visible=True,
        )

        # Slug should be stored properly
        self.assertLessEqual(len(plan.slug), 100)
        self.assertTrue(plan.slug.startswith("plan-with-many-words"))


class SlugMigrationDataIntegrityTestCase(TestCase):
    """Test data integrity during slug migration."""

    def setUp(self):
        """Set up test data."""
        # Clear any existing plans to avoid conflicts
        Plan.objects.all().delete()

    def test_migration_preserves_all_plan_data(self):
        """Test that migration preserves all existing plan data."""
        # Create a plan with all fields populated
        original_data = {
            "name": "Complete Plan",
            "slug": "complete-plan",
            "description": "Plan with all fields",
            "available": True,
            "visible": False,
            "default": True,
        }

        plan = Plan.objects.create(**original_data)
        original_id = plan.id

        # Verify all data is preserved after save
        plan.refresh_from_db()

        self.assertEqual(plan.id, original_id)
        self.assertEqual(plan.name, original_data["name"])
        self.assertEqual(plan.slug, original_data["slug"])
        self.assertEqual(plan.description, original_data["description"])
        self.assertEqual(plan.available, original_data["available"])
        self.assertEqual(plan.visible, original_data["visible"])
        self.assertEqual(plan.default, original_data["default"])

    def test_migration_with_related_objects(self):
        """Test that migration works correctly with related objects."""
        # Create a plan that might have related objects
        plan = Plan.objects.create(
            name="Plan with Relations",
            slug="plan-with-relations",
            description="Plan that might have quotas and pricing",
            available=True,
            visible=True,
        )

        # Verify the plan exists and has correct slug
        self.assertTrue(Plan.objects.filter(id=plan.id).exists())
        self.assertEqual(plan.slug, "plan-with-relations")

        # Verify we can still access the plan by its relationships
        # (This tests that foreign key relationships aren't broken)
        retrieved_plan = Plan.objects.get(id=plan.id)
        self.assertEqual(retrieved_plan.slug, plan.slug)



# Generated by Django 2.0.6 on 2018-06-12 12:30

from decimal import Decimal

import django.db.models.deletion
import django_countries.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BillingInfo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "tax_number",
                    models.CharField(
                        blank=True, db_index=True, max_length=200, verbose_name="VAT ID"
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True, max_length=200, verbose_name="name"
                    ),
                ),
                ("street", models.CharField(max_length=200, verbose_name="street")),
                ("zipcode", models.CharField(max_length=200, verbose_name="zip code")),
                ("city", models.CharField(max_length=200, verbose_name="city")),
                (
                    "country",
                    django_countries.fields.CountryField(
                        max_length=2, verbose_name="country"
                    ),
                ),
                (
                    "shipping_name",
                    models.CharField(
                        blank=True,
                        help_text="optional",
                        max_length=200,
                        verbose_name="name (shipping)",
                    ),
                ),
                (
                    "shipping_street",
                    models.CharField(
                        blank=True,
                        help_text="optional",
                        max_length=200,
                        verbose_name="street (shipping)",
                    ),
                ),
                (
                    "shipping_zipcode",
                    models.CharField(
                        blank=True,
                        help_text="optional",
                        max_length=200,
                        verbose_name="zip code (shipping)",
                    ),
                ),
                (
                    "shipping_city",
                    models.CharField(
                        blank=True,
                        help_text="optional",
                        max_length=200,
                        verbose_name="city (shipping)",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "Billing info",
                "verbose_name_plural": "Billing infos",
            },
        ),
        migrations.CreateModel(
            name="Invoice",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("number", models.IntegerField(db_index=True)),
                ("full_number", models.CharField(max_length=200)),
                (
                    "type",
                    models.IntegerField(
                        choices=[
                            (1, "Invoice"),
                            (2, "Invoice Duplicate"),
                            (3, "Order confirmation"),
                        ],
                        db_index=True,
                        default=1,
                    ),
                ),
                ("issued", models.DateField(db_index=True)),
                (
                    "issued_duplicate",
                    models.DateField(blank=True, db_index=True, null=True),
                ),
                (
                    "selling_date",
                    models.DateField(blank=True, db_index=True, null=True),
                ),
                ("payment_date", models.DateField(db_index=True)),
                ("unit_price_net", models.DecimalField(decimal_places=2, max_digits=7)),
                ("quantity", models.IntegerField(default=1)),
                ("total_net", models.DecimalField(decimal_places=2, max_digits=7)),
                ("total", models.DecimalField(decimal_places=2, max_digits=7)),
                ("tax_total", models.DecimalField(decimal_places=2, max_digits=7)),
                (
                    "tax",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        max_digits=4,
                        null=True,
                    ),
                ),
                (
                    "rebate",
                    models.DecimalField(
                        decimal_places=2, default=Decimal("0"), max_digits=4
                    ),
                ),
                ("currency", models.CharField(default="EUR", max_length=3)),
                ("item_description", models.CharField(max_length=200)),
                ("buyer_name", models.CharField(max_length=200, verbose_name="Name")),
                (
                    "buyer_street",
                    models.CharField(max_length=200, verbose_name="Street"),
                ),
                (
                    "buyer_zipcode",
                    models.CharField(max_length=200, verbose_name="Zip code"),
                ),
                ("buyer_city", models.CharField(max_length=200, verbose_name="City")),
                (
                    "buyer_country",
                    django_countries.fields.CountryField(
                        default="PL", max_length=2, verbose_name="Country"
                    ),
                ),
                (
                    "buyer_tax_number",
                    models.CharField(
                        blank=True, max_length=200, verbose_name="TAX/VAT number"
                    ),
                ),
                (
                    "shipping_name",
                    models.CharField(max_length=200, verbose_name="Name"),
                ),
                (
                    "shipping_street",
                    models.CharField(max_length=200, verbose_name="Street"),
                ),
                (
                    "shipping_zipcode",
                    models.CharField(max_length=200, verbose_name="Zip code"),
                ),
                (
                    "shipping_city",
                    models.CharField(max_length=200, verbose_name="City"),
                ),
                (
                    "shipping_country",
                    django_countries.fields.CountryField(
                        default="PL", max_length=2, verbose_name="Country"
                    ),
                ),
                ("require_shipment", models.BooleanField(db_index=True, default=False)),
                ("issuer_name", models.CharField(max_length=200, verbose_name="Name")),
                (
                    "issuer_street",
                    models.CharField(max_length=200, verbose_name="Street"),
                ),
                (
                    "issuer_zipcode",
                    models.CharField(max_length=200, verbose_name="Zip code"),
                ),
                ("issuer_city", models.CharField(max_length=200, verbose_name="City")),
                (
                    "issuer_country",
                    django_countries.fields.CountryField(
                        default="PL", max_length=2, verbose_name="Country"
                    ),
                ),
                (
                    "issuer_tax_number",
                    models.CharField(
                        blank=True, max_length=200, verbose_name="TAX/VAT number"
                    ),
                ),
            ],
            options={
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
            },
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("flat_name", models.CharField(blank=True, max_length=200, null=True)),
                (
                    "created",
                    models.DateTimeField(db_index=True, verbose_name="created"),
                ),
                (
                    "completed",
                    models.DateTimeField(
                        blank=True, db_index=True, null=True, verbose_name="completed"
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        db_index=True,
                        decimal_places=2,
                        max_digits=7,
                        verbose_name="amount",
                    ),
                ),
                (
                    "tax",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        max_digits=4,
                        null=True,
                        verbose_name="tax",
                    ),
                ),
                (
                    "currency",
                    models.CharField(
                        default="EUR", max_length=3, verbose_name="currency"
                    ),
                ),
                (
                    "status",
                    models.IntegerField(
                        choices=[
                            (1, "new"),
                            (2, "completed"),
                            (3, "not valid"),
                            (4, "canceled"),
                            (5, "returned"),
                        ],
                        default=1,
                        verbose_name="status",
                    ),
                ),
            ],
            options={
                "verbose_name": "Order",
                "verbose_name_plural": "Orders",
                "ordering": ("-created",),
            },
        ),
        migrations.CreateModel(
            name="Plan",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("order", models.PositiveIntegerField(db_index=True, editable=False)),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                ("default", models.BooleanField(db_index=True, default=False)),
                (
                    "available",
                    models.BooleanField(
                        db_index=True,
                        default=False,
                        help_text="Is still available for purchase",
                        verbose_name="available",
                    ),
                ),
                (
                    "visible",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="Is visible in current offer",
                        verbose_name="visible",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(db_index=True, verbose_name="created"),
                ),
                (
                    "url",
                    models.URLField(
                        blank=True,
                        help_text="Optional link to page with more information (for clickable pricing table headers)",
                    ),
                ),
                (
                    "customized",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="customized",
                    ),
                ),
            ],
            options={
                "verbose_name": "Plan",
                "verbose_name_plural": "Plans",
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="PlanPricing",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "price",
                    models.DecimalField(db_index=True, decimal_places=2, max_digits=7),
                ),
                (
                    "plan",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="plans.Plan"
                    ),
                ),
            ],
            options={
                "verbose_name": "Plan pricing",
                "verbose_name_plural": "Plans pricings",
                "ordering": ("pricing__period",),
            },
        ),
        migrations.CreateModel(
            name="PlanQuota",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.IntegerField(blank=True, default=1, null=True)),
                (
                    "plan",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="plans.Plan"
                    ),
                ),
            ],
            options={
                "verbose_name": "Plan quota",
                "verbose_name_plural": "Plans quotas",
            },
        ),
        migrations.CreateModel(
            name="Pricing",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                (
                    "period",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        default=30,
                        null=True,
                        verbose_name="period",
                    ),
                ),
                (
                    "url",
                    models.URLField(
                        blank=True,
                        help_text="Optional link to page with more information (for clickable pricing table headers)",
                    ),
                ),
            ],
            options={
                "verbose_name": "Pricing",
                "verbose_name_plural": "Pricings",
                "ordering": ("period",),
            },
        ),
        migrations.CreateModel(
            name="Quota",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("order", models.PositiveIntegerField(db_index=True, editable=False)),
                (
                    "codename",
                    models.CharField(
                        db_index=True,
                        max_length=50,
                        unique=True,
                        verbose_name="codename",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                (
                    "unit",
                    models.CharField(blank=True, max_length=100, verbose_name="unit"),
                ),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "is_boolean",
                    models.BooleanField(default=False, verbose_name="is boolean"),
                ),
                (
                    "url",
                    models.CharField(
                        blank=True,
                        help_text="Optional link to page with more information (for clickable pricing table headers)",
                        max_length=200,
                    ),
                ),
            ],
            options={
                "verbose_name": "Quota",
                "verbose_name_plural": "Quotas",
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="UserPlan",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "expire",
                    models.DateField(
                        blank=True,
                        db_index=True,
                        default=None,
                        null=True,
                        verbose_name="expire",
                    ),
                ),
                (
                    "active",
                    models.BooleanField(
                        db_index=True, default=True, verbose_name="active"
                    ),
                ),
                (
                    "plan",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="plans.Plan",
                        verbose_name="plan",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "User plan",
                "verbose_name_plural": "Users plans",
            },
        ),
        migrations.AddField(
            model_name="planquota",
            name="quota",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="plans.Quota"
            ),
        ),
        migrations.AddField(
            model_name="planpricing",
            name="pricing",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="plans.Pricing"
            ),
        ),
        migrations.AddField(
            model_name="plan",
            name="quotas",
            field=models.ManyToManyField(through="plans.PlanQuota", to="plans.Quota"),
        ),
        migrations.AddField(
            model_name="order",
            name="plan",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="plan_order",
                to="plans.Plan",
                verbose_name="plan",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="pricing",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="plans.Pricing",
                verbose_name="pricing",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="order",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="plans.Order",
                verbose_name="order",
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
    ]



# Generated by Django 5.2.2 on 2025-06-18 09:02

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("plans", "0015_recurringuserplan_last_renewal_attempt"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoice",
            name="cancellation_reason",
            field=models.TextField(blank=True, verbose_name="Cancellation reason"),
        ),
        migrations.AddField(
            model_name="invoice",
            name="credit_note_for",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="issued_credit_notes",
                to=settings.PLANS_INVOICE_MODEL,
                verbose_name="Credit note for",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="type",
            field=models.IntegerField(
                choices=[
                    (1, "Invoice"),
                    (2, "Invoice Duplicate"),
                    (3, "Order confirmation"),
                    (4, "Credit note"),
                ],
                db_index=True,
                default=1,
            ),
        ),
    ]



# TODO use django-conf
from django.conf import settings

TAX = getattr(settings, "PLANS_TAX", None)
TAXATION_POLICY = getattr(
    settings, "PLANS_TAXATION_POLICY", "plans.taxation.TAXATION_POLICY"
)
APP_VERBOSE_NAME = getattr(settings, "PLANS_APP_VERBOSE_NAME", "plans")



"""
Comprehensive tests for the entire migration chain to ensure migrations work
on both blank databases and existing databases.

Tests for issue #192 and general migration robustness.
"""

from django.apps import apps
from django.contrib.auth import get_user_model
from django.test import TestCase

User = get_user_model()


class MigrationChainTestCase(TestCase):
    """Test migration logic without actually running migrations on the test database."""

    def setUp(self):
        """Set up test data for migration logic testing."""
        # Test the migration logic without actually manipulating schema
        pass

    def test_migration_0004_logic_isolated(self):
        """Test the migration 0004 logic in isolation without schema changes."""
        # This tests the actual logic that migration 0004 uses
        # without requiring schema manipulation

        # Create test data
        user = User.objects.create_user(
            username="migration_test_user",
            email="test@migration.com",
            password="testpass",
        )

        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Create a default plan
        plan = Plan.objects.create(
            name="Migration Test Plan",
            description="Plan for testing migration logic",
            default=True,
            available=True,
            visible=True,
        )

        # Ensure no user plan exists initially
        UserPlan.objects.filter(user=user).delete()

        # Test the migration 0004 logic (using current models, but same logic)
        default_plan = Plan.objects.filter(default=True).first()
        self.assertIsNotNone(default_plan, "Should have a default plan")

        users_without_plans = User.objects.filter(userplan=None)
        self.assertGreaterEqual(
            users_without_plans.count(), 1, "Should have users without plans"
        )

        # Apply the migration logic
        for user_without_plan in users_without_plans:
            UserPlan.objects.create(
                user=user_without_plan,
                plan=default_plan,
                active=False,
                expire=None,
            )

        # Verify results
        user_plan = UserPlan.objects.get(user=user)
        self.assertEqual(user_plan.plan, plan)
        self.assertFalse(user_plan.active)
        self.assertIsNone(user_plan.expire)

    def test_migration_handles_no_default_plan_scenario(self):
        """Test that migration logic handles missing default plan gracefully."""
        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Create a user
        user = User.objects.create_user(
            username="no_default_user", email="nodefault@test.com", password="testpass"
        )

        # Ensure no default plan exists
        Plan.objects.filter(default=True).delete()

        # Test migration logic with no default plan
        default_plan = Plan.objects.filter(default=True).first()
        self.assertIsNone(default_plan, "Should have no default plan")

        # Migration should handle this gracefully (skip user plan creation)
        if default_plan is None:
            # This is what the migration does - it returns early
            initial_count = UserPlan.objects.filter(user=user).count()
            # No user plans should be created
            final_count = UserPlan.objects.filter(user=user).count()
            self.assertEqual(
                initial_count,
                final_count,
                "No user plans should be created without default plan",
            )


class MigrationDataIntegrityTestCase(TestCase):
    """Test that migrations preserve data integrity."""

    def test_user_plan_creation_with_existing_users(self):
        """Test that migration 0004 creates user plans for existing users."""
        # Create users before running the migration logic
        user1 = User.objects.create_user(
            username="existing_user1", email="user1@test.com", password="testpass"
        )
        user2 = User.objects.create_user(
            username="existing_user2", email="user2@test.com", password="testpass"
        )

        # Create a default plan
        Plan = apps.get_model("plans", "Plan")
        plan = Plan.objects.create(
            name="Default Plan",
            description="Default plan for existing users",
            default=True,
            available=True,
            visible=True,
        )

        # Simulate the migration 0004 logic
        UserPlan = apps.get_model("plans", "UserPlan")

        # Delete any existing user plans
        UserPlan.objects.all().delete()

        # Run the migration logic
        default_plan = Plan.objects.filter(default=True).first()
        self.assertIsNotNone(default_plan)

        users_without_plans = User.objects.filter(userplan=None)
        self.assertGreaterEqual(users_without_plans.count(), 2)

        for user in users_without_plans:
            UserPlan.objects.create(
                user=user,
                plan=default_plan,
                active=False,
                expire=None,
            )

        # Verify both users got user plans
        self.assertTrue(UserPlan.objects.filter(user=user1).exists())
        self.assertTrue(UserPlan.objects.filter(user=user2).exists())

        user1_plan = UserPlan.objects.get(user=user1)
        user2_plan = UserPlan.objects.get(user=user2)

        self.assertEqual(user1_plan.plan, plan)
        self.assertEqual(user2_plan.plan, plan)
        self.assertFalse(user1_plan.active)
        self.assertFalse(user2_plan.active)

    def test_migration_with_existing_default_plan(self):
        """Test migration behavior when a default plan already exists."""
        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Ensure no default plan exists initially
        Plan.objects.filter(default=True).delete()

        # Create a default plan
        plan = Plan.objects.create(
            name="Existing Default Plan",
            description="Plan that already exists as default",
            default=True,
            available=True,
            visible=True,
        )

        user = User.objects.create_user(
            username="existing_default_user",
            email="existing@test.com",
            password="testpass",
        )

        # Migration should use the existing default plan
        default_plan = Plan.objects.filter(default=True).first()
        self.assertEqual(default_plan, plan)

        # Create user plan using migration logic
        UserPlan.objects.filter(user=user).delete()
        UserPlan.objects.create(
            user=user,
            plan=default_plan,
            active=False,
            expire=None,
        )

        user_plan = UserPlan.objects.get(user=user)
        self.assertEqual(user_plan.plan, plan)

    def test_migration_with_no_users(self):
        """Test that migration works correctly when no users exist."""
        Plan = apps.get_model("plans", "Plan")
        UserPlan = apps.get_model("plans", "UserPlan")

        # Create a default plan
        Plan.objects.create(
            name="No Users Plan",
            description="Plan when no users exist",
            default=True,
            available=True,
            visible=True,
        )

        # Ensure no users exist
        User.objects.all().delete()

        # Run migration logic
        default_plan = Plan.objects.filter(default=True).first()
        self.assertIsNotNone(default_plan)

        users_without_plans = User.objects.filter(userplan=None)
        self.assertEqual(users_without_plans.count(), 0)

        # Should not create any user plans
        initial_count = UserPlan.objects.count()
        for user in users_without_plans:
            UserPlan.objects.create(
                user=user,
                plan=default_plan,
                active=False,
                expire=None,
            )
        final_count = UserPlan.objects.count()

        self.assertEqual(initial_count, final_count)



import logging
from datetime import datetime
from decimal import Decimal
from typing import Optional

from django.core.cache import cache
from requests.exceptions import ConnectionError, Timeout
from zeep import Client
from zeep.exceptions import Fault, TransportError, XMLSyntaxError

logger = logging.getLogger("plans.taxation.tedb")


class TEDBClient:
    """
    Client for European Commission's TEDB SOAP web service.
    Provides real-time VAT rates with caching and fallback mechanisms.
    """

    WSDL_URL = "https://ec.europa.eu/taxation_customs/tedb/ws/VatRetrievalService.wsdl"
    CACHE_TIMEOUT = 3600 * 24  # 24 hours
    CACHE_KEY_PREFIX = "tedb_vat_rate"

    def __init__(self):
        self.client = None
        self._initialize_client()

    def _initialize_client(self):
        """Initialize SOAP client with error handling."""
        try:
            self.client = Client(self.WSDL_URL)
            logger.info("TEDB SOAP client initialized successfully")
        except (ConnectionError, Timeout, TransportError, Fault, XMLSyntaxError) as e:
            logger.error(f"Failed to initialize TEDB SOAP client: {e}")
            self.client = None

    def _get_cache_key(self, country_code: str, date: str = None) -> str:
        """Generate cache key for VAT rate."""
        if not date:
            date = datetime.now().strftime("%Y-%m-%d")
        return f"{self.CACHE_KEY_PREFIX}_{country_code}_{date}"

    def get_vat_rate(self, country_code: str) -> Optional[Decimal]:
        """
        Retrieve VAT rate for a country from TEDB service.

        Args:
            country_code: ISO 2-letter country code (e.g., 'DE', 'FR')

        Returns:
            Decimal VAT rate or None if not found/error
        """
        date = datetime.now().strftime("%Y-%m-%d")

        # Check cache first
        cache_key = self._get_cache_key(country_code, date)
        cached_rate = cache.get(cache_key)
        if cached_rate is not None:
            logger.debug(f"Retrieved cached VAT rate for {country_code}: {cached_rate}")
            return cached_rate

        # Try TEDB service
        if self.client:
            try:
                logger.info(f"Retrieving VAT rate from TEDB for {country_code}")

                response = self.client.service.retrieveVatRates(
                    memberStates={"isoCode": [country_code]}, situationOn=date
                )

                if hasattr(response, "vatRateResults") and response.vatRateResults:
                    # Collect all standard/default VAT rates
                    # Some countries have multiple rates (e.g., Spain has Canary Islands)
                    standard_rates = []

                    for vat_rate in response.vatRateResults:
                        if (
                            hasattr(vat_rate, "memberState")
                            and vat_rate.memberState == country_code
                            and hasattr(vat_rate, "rate")
                            and hasattr(vat_rate, "type")
                            and vat_rate.type == "STANDARD"
                        ):
                            rate_info = vat_rate.rate
                            if hasattr(rate_info, "value") and hasattr(
                                rate_info, "type"
                            ):
                                if rate_info.type == "DEFAULT":
                                    # Store rate with metadata for filtering
                                    comment = (
                                        vat_rate.comment
                                        if hasattr(vat_rate, "comment")
                                        else None
                                    )
                                    standard_rates.append(
                                        {"value": rate_info.value, "comment": comment}
                                    )

                    if standard_rates:
                        # Prefer rate without comment (general/mainland rate)
                        # Filter out special regions (e.g., Canary Islands)
                        general_rates = [r for r in standard_rates if not r["comment"]]

                        if general_rates:
                            # Use the general rate
                            selected_rate = general_rates[0]["value"]
                        else:
                            # If all have comments, use the highest rate
                            selected_rate = max(r["value"] for r in standard_rates)
                            logger.warning(
                                f"Multiple regional rates for {country_code}, using highest: {selected_rate}%"
                            )

                        # Convert to Decimal and normalize
                        raw_decimal = Decimal(str(selected_rate))
                        rate = (
                            raw_decimal.quantize(Decimal("1"))
                            if raw_decimal % 1 == 0
                            else raw_decimal
                        )

                        # Cache the result
                        cache.set(cache_key, rate, self.CACHE_TIMEOUT)
                        logger.info(
                            f"Retrieved standard VAT rate from TEDB for {country_code}: {rate}%"
                        )
                        return rate
            except (Fault, TransportError, ConnectionError, Timeout) as e:
                logger.warning(f"TEDB service error for {country_code}: {e}")

        logger.warning(f"Could not retrieve VAT rate from TEDB for {country_code}")
        return None



import logging

from django.core.management import BaseCommand

from plans import tasks


class Command(BaseCommand):
    help = "Autorenew accounts and with recurring payments"

    def add_arguments(self, parser):
        parser.add_argument(
            "--providers",
            nargs="+",
            dest="providers",
            help="Renew only accounts with this providers",
        )
        parser.add_argument(
            "--throttle",
            type=int,
            dest="throttle",
            help="Throttle seconds between renewals",
        )
        parser.add_argument(
            "--catch-exceptions",
            action="store_true",
            dest="catch_exceptions",
            help="Catch exceptions during renewal and log them",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry_run",
            help="Dry run, do not change any data",
        )

    def handle(self, *args, **options):  # pragma: no cover
        logger = logging.getLogger("plans.tasks")
        handler = logging.StreamHandler(self.stdout)
        logger.addHandler(handler)
        verbosity = options.get("verbosity")
        if verbosity == 0:
            logger.setLevel(logging.WARNING)
        elif verbosity == 1:
            logger.setLevel(logging.INFO)
        else:  # verbosity > 1
            logger.setLevel(logging.DEBUG)

        try:
            providers = options.get("providers")
            dry_run = options.get("dry_run")
            self.stdout.write("Starting renewal")
            if dry_run:
                self.stdout.write("DRY RUN active")
            throttle_seconds = options.get("throttle")
            catch_exceptions = options.get("catch_exceptions")
            renewed_accounts = tasks.autorenew_account(
                providers,
                throttle_seconds=throttle_seconds,
                catch_exceptions=catch_exceptions,
                dry_run=dry_run,
            )
            if renewed_accounts:
                if dry_run:
                    self.stdout.write(
                        f"{len(renewed_accounts)} accounts would be submitted to renewal:"
                    )
                else:
                    self.stdout.write(
                        f"{len(renewed_accounts)} accounts submitted to renewal:"
                    )
                for a in renewed_accounts:
                    self.stdout.write(
                        f"\t{a.userplan.recurring.payment_provider:<30}{a.email:<40}{str(a).strip():<40}"
                        f"{a.userplan.expire}\t{a.userplan.active}"
                    )
            else:
                self.stdout.write("No accounts autorenewed")
        finally:
            logger.removeHandler(handler)



# Generated by Django 3.0.5 on 2020-04-15 07:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0004_create_user_plans"),
    ]

    operations = [
        migrations.AddField(
            model_name="planpricing",
            name="has_automatic_renewal",
            field=models.BooleanField(
                default=False,
                help_text="Use automatic renewal if possible?",
                verbose_name="has automatic renewal",
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="order",
            field=models.PositiveIntegerField(
                db_index=True, editable=False, verbose_name="order"
            ),
        ),
        migrations.AlterField(
            model_name="quota",
            name="order",
            field=models.PositiveIntegerField(
                db_index=True, editable=False, verbose_name="order"
            ),
        ),
        migrations.CreateModel(
            name="RecurringUserPlan",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "token",
                    models.CharField(
                        blank=True,
                        default=None,
                        help_text="Token, that will be used for payment renewal. Depends on used payment provider",
                        max_length=255,
                        null=True,
                        verbose_name="recurring token",
                    ),
                ),
                (
                    "payment_provider",
                    models.CharField(
                        blank=True,
                        default=None,
                        help_text="Provider, that will be used for payment renewal",
                        max_length=255,
                        null=True,
                        verbose_name="payment provider",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        max_digits=7,
                        null=True,
                        verbose_name="amount",
                    ),
                ),
                (
                    "tax",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        max_digits=4,
                        null=True,
                        verbose_name="tax",
                    ),
                ),
                ("currency", models.CharField(max_length=3, verbose_name="currency")),
                (
                    "has_automatic_renewal",
                    models.BooleanField(
                        default=False,
                        help_text="Automatic renewal is enabled for associated plan. If False, the plan renewal can be still initiated by user.",
                        verbose_name="has automatic plan renewal",
                    ),
                ),
                ("card_expire_year", models.IntegerField(blank=True, null=True)),
                ("card_expire_month", models.IntegerField(blank=True, null=True)),
                (
                    "pricing",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Recurring pricing",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="plans.Pricing",
                    ),
                ),
                (
                    "user_plan",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="recurring",
                        to="plans.UserPlan",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-04 13:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0005_recurring_payments"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="planpricing",
            options={
                "ordering": ("order", "pricing__period"),
                "verbose_name": "Plan pricing",
                "verbose_name_plural": "Plans pricings",
            },
        ),
        migrations.AddField(
            model_name="planpricing",
            name="order",
            field=models.IntegerField(default=0),
        ),
    ]



# Generated by Django 3.2.11 on 2022-02-08 12:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("plans", "0010_auto_20220113_1317"),
    ]

    operations = [
        migrations.AddField(
            model_name="billinginfo",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="billinginfo",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="invoice",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="invoice",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="order",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="plan",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="planpricing",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="planpricing",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="planquota",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="planquota",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="pricing",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="pricing",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="quota",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="quota",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="recurringuserplan",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="recurringuserplan",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AddField(
            model_name="userplan",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True,
                db_index=True,
                default=None,
                null=True,
                verbose_name="created",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="userplan",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="shipping_street",
            field=models.CharField(blank=True, max_length=200, verbose_name="Street"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="shipping_zipcode",
            field=models.CharField(blank=True, max_length=200, verbose_name="Zip code"),
        ),
        migrations.AlterField(
            model_name="order",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, null=True, verbose_name="created"
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="created",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, null=True, verbose_name="created"
            ),
        ),
    ]



from decimal import Decimal

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from plans.importer import import_name


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def get_country_code(request):
    if getattr(settings, "PLANS_GET_COUNTRY_FROM_IP", False):
        try:
            from geolite2 import geolite2

            reader = geolite2.reader()
            ip_address = get_client_ip(request)
            ip_info = reader.get(ip_address)
        except ModuleNotFoundError:
            ip_info = None

        if ip_info and "country" in ip_info:
            country_code = ip_info["country"]["iso_code"]
            return country_code
    return getattr(settings, "PLANS_DEFAULT_COUNTRY", None)


def get_currency():
    CURRENCY = getattr(settings, "PLANS_CURRENCY", "")
    if len(CURRENCY) != 3:
        raise ImproperlyConfigured(
            "PLANS_CURRENCY should be configured as 3-letter currency code."
        )
    return CURRENCY


def country_code_transform(country_code):
    """Transform country code to the code used by VIES"""
    transform_dict = {
        "GR": "EL",
    }
    return transform_dict.get(country_code, country_code)


def calculate_tax_rate(tax_number, country_code, request=None):
    taxation_policy = getattr(settings, "PLANS_TAXATION_POLICY", None)
    if not taxation_policy:
        raise ImproperlyConfigured("PLANS_TAXATION_POLICY is not set")
    taxation_policy = import_name(taxation_policy)
    tax, request_successful = taxation_policy.get_tax_rate(
        tax_number, country_code, request
    )
    if request_successful and request:
        TaxCacheService.cache_tax_rate(request, tax, tax_number, country_code)
    return Decimal(tax) if tax is not None else None, request_successful


def get_tax_rate(country_code, tax_number, request=None):
    """Get tax rate for given country and tax number
    1. Try to get tax rate from cache
    2. If not in cache, calculate it (and possibly cache it)

    Returns tax rate and if the request was successful (False means default tax rate was used)
    """
    if request:
        try:
            tax_from_cache = TaxCacheService.get_tax_rate(
                request, tax_number, country_code
            )
            return tax_from_cache, True
        except KeyError:
            pass

    tax, request_successful = calculate_tax_rate(tax_number, country_code, request)
    return tax, request_successful


class TaxCacheService:
    @classmethod
    def get_cache_key(cls, tax_number, country):
        return "tax_%s_%s" % (tax_number, country)

    @classmethod
    def cache_tax_rate(cls, request, tax, tax_number, country):
        request.session[cls.get_cache_key(tax_number, country)] = str(tax)

    @classmethod
    def get_tax_rate(cls, request, tax_number, country):
        key = cls.get_cache_key(tax_number, country)
        if key not in request.session:
            raise KeyError(
                f"Tax rate for {tax_number} and {country} not found in cache"
            )
        raw = request.session[key]
        if raw == "None":
            return None
        return Decimal(raw)
